"""Module with generic usable utility functions for using OWS services."""

import logging
import math
import random
import time
import warnings
from concurrent import futures
from pathlib import Path
from typing import Optional, Union

import geofileops as gfo
import geopandas as gpd
import numpy as np
import owslib
import owslib.util
import owslib.wms
import pycron
import pygeoops
import pyproj
import rasterio as rio
import rasterio.enums
import rasterio.errors as rio_errors
import urllib3
from rasterio import (
    profiles as rio_profiles,
    transform as rio_transform,
    windows as rio_windows,
)
from rasterio._err import CPLE_AppDefinedError

from . import progress_util

FORMAT_GEOTIFF = "image/geotiff"
FORMAT_GEOTIFF_DRIVER = "Gtiff"
FORMAT_GEOTIFF_EXT = ".tif"
FORMAT_GEOTIFF_EXT_WORLD = ".tfw"

FORMAT_TIFF = "image/tiff"
FORMAT_TIFF_DRIVER = "Gtiff"
FORMAT_TIFF_EXT = ".tif"
FORMAT_TIFF_EXT_WORLD = ".tfw"

FORMAT_JPEG = "image/jpeg"
FORMAT_JPEG_DRIVER = "JPEG"
FORMAT_JPEG_EXT = ".jpg"
FORMAT_JPEG_EXT_WORLD = ".jgw"

FORMAT_PNG = "image/png"
FORMAT_PNG_DRIVER = "PNG"
FORMAT_PNG_EXT = ".png"
FORMAT_PNG_EXT_WORLD = ".pgw"

# Get a logger...
logger = logging.getLogger(__name__)


class WMSLayerSource:
    """Properties of a WMS layer source."""

    def __init__(
        self,
        wms_server_url: str,
        layernames: list[str],
        layerstyles: Optional[list[str]] = None,
        bands: Optional[list[int]] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        wms_version: str = "1.3.0",
        wms_ignore_capabilities_url: bool = False,
        random_sleep: int = 0,
        wms_service: Union[
            owslib.wms.wms111.WebMapService_1_1_1,
            owslib.wms.wms130.WebMapService_1_3_0,
            None,
        ] = None,
    ):
        """Constructor of WMSLayerSource.

        Args:
            wms_server_url (str): _description_
            layernames (list[str]): _description_
            layerstyles (Optional[list[str]], optional): _description_.
                Defaults to None.
            bands (Optional[list[int]], optional): _description_. Defaults to None.
            username (str, optional): username to logon with. Defaults to None.
            password (str, optional): password to logon with. Defaults to None.
            wms_version (str, optional): _description_. Defaults to "1.3.0".
            wms_ignore_capabilities_url (bool, optional): _description_.
                Defaults to False.
            random_sleep (int, optional): _description_. Defaults to 0.
            wms_service: The WMS service. Defaults to None.
        """
        self.wms_server_url = wms_server_url
        self.wms_version = wms_version
        self.wms_ignore_capabilities_url = wms_ignore_capabilities_url
        self.layernames = layernames
        self.layerstyles = layerstyles
        self.bands = bands
        self.username = username
        self.password = password
        self.random_sleep = random_sleep
        self.wms_service = wms_service


class FileLayerSource:
    """Properties of a file layer source."""

    def __init__(
        self,
        path: Union[str, Path],
        layernames: list[str],
        bands: Optional[list[int]] = None,
    ):
        """Contructor for FileLayerSource.

        Args:
            path (Union[str, Path]): Path to the layer.
            layernames (list[str]): list of layer names.
            bands (Optional[list[int]], optional): list of bands. Defaults to None.
        """
        self.path = Path(path)
        self.layernames = layernames
        self.bands = bands


def get_images_for_grid(
    layersources: list[Union[FileLayerSource, WMSLayerSource]],
    output_image_dir: Path,
    crs: Union[str, pyproj.CRS],
    image_gen_bbox: Optional[tuple[float, float, float, float]] = None,
    image_gen_roi_filepath: Optional[Path] = None,
    grid_xmin: float = 0.0,
    grid_ymin: float = 0.0,
    image_crs_pixel_x_size: float = 0.25,
    image_crs_pixel_y_size: float = 0.25,
    image_pixel_width: int = 1024,
    image_pixel_height: int = 1024,
    image_pixels_ignore_border: int = 0,
    nb_concurrent_calls: int = 1,
    cron_schedule: Optional[str] = None,
    image_format: str = FORMAT_GEOTIFF,
    image_format_save: Optional[str] = None,
    tiff_compress: str = "lzw",
    transparent: bool = False,
    pixels_overlap: int = 0,
    nb_images_to_skip: int = 0,
    max_nb_images: int = -1,
    ssl_verify: Union[bool, str] = True,
    force: bool = False,
):
    """Loads all images in a grid from a WMS service.

    Args:
        layersources (list[dict]): Layer sources to get images from. Multiple
            sources can be specified to create a combined image, eg. use band
            1 of a WMS service with band 2 and 3 of another one.
        output_image_dir (Path): Directory to save the images to.
        crs (Union[str, pyproj.CRS]): The crs of the source and destination images.
        image_gen_bbox (tuple[float, float, float, float], optional): bbox of the roi to
            request/save images for. Defaults to None.
        image_gen_roi_filepath (Optional[Path], optional): File with the roi
            where images should be requested/saved for. Defaults to None.
        grid_xmin (float, optional): xmin for the grid to be used.
            Defaults to 0.0.
        grid_ymin (float, optional): ymin for the grid to be used.
            Defaults to 0.0.
        image_crs_pixel_x_size (float, optional): [description]. Defaults to 0.25.
        image_crs_pixel_y_size (float, optional): [description]. Defaults to 0.25.
        image_pixel_width (int, optional): [description]. Defaults to 1024.
        image_pixel_height (int, optional): [description]. Defaults to 1024.
        image_pixels_ignore_border (int, optional): [description]. Defaults to 0.
        nb_concurrent_calls (int, optional): Number of images to treat in
            parallel. Will increase the load on the WMS server! Defaults to 1.
        cron_schedule (str, optional): [description]. Defaults to None.
        image_format (str, optional): The image format to get. Defaults to
            FORMAT_GEOTIFF.
        image_format_save (str, optional): The image format to save to.
            Defaults to None.
        tiff_compress (str, optional): [description]. Defaults to 'lzw'.
        transparent (bool, optional): [description]. Defaults to False.
        pixels_overlap (int, optional): [description]. Defaults to 0.
        nb_images_to_skip (int, optional): [description]. Defaults to 0.
        max_nb_images (int, optional): [description]. Defaults to -1.
        ssl_verify (bool or str, optional): True to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to True.
        force (bool, optional): [description]. Defaults to False.
    """
    # Init
    if image_format_save is None:
        image_format_save = image_format
    if not isinstance(crs, pyproj.CRS):
        crs = pyproj.CRS(crs)

    crs_width = math.fabs(
        image_pixel_width * image_crs_pixel_x_size
    )  # tile width in units of crs => 500 m
    crs_height = math.fabs(
        image_pixel_height * image_crs_pixel_y_size
    )  # tile height in units of crs => 500 m
    if cron_schedule is not None and cron_schedule != "":
        logger.info(
            "A cron_schedule was specified, so the download will only proceed in the "
            f"specified time range: {cron_schedule}"
        )

    # Read the region of interest file if provided
    grid_bbox = image_gen_bbox
    roi_gdf = None
    if image_gen_roi_filepath is not None:
        # Read roi
        logger.info(f"Read + optimize region of interest from {image_gen_roi_filepath}")
        roi_gdf = gpd.read_file(str(image_gen_roi_filepath))

        # If the generate_window not specified, use bounds of the roi
        if grid_bbox is None:
            grid_bbox = roi_gdf.geometry.total_bounds

    # If there is still no image_gen_bbox, stop.
    if grid_bbox is None:
        raise Exception(
            "Either image_gen_bbox or an image_gen_roi_filepath should be specified."
        )

    # Make grid_bounds compatible with the grid
    grid_bbox = align_bbox_to_grid(
        bbox=grid_bbox,
        grid_xmin=grid_xmin,
        grid_ymin=grid_ymin,
        pixel_size_x=crs_width,
        pixel_size_y=crs_height,
        log_level=logging.WARNING,
    )

    # Create the output dir if it doesn't exist yet
    output_image_dir.mkdir(parents=True, exist_ok=True)

    # Create a grid of the images that need to be downloaded
    tiles_to_download_gdf = gpd.GeoDataFrame(
        geometry=pygeoops.create_grid3(grid_bbox, width=crs_width, height=crs_height),
        crs=crs,
    )

    # If an roi is defined, filter the split it using a grid as large objects are small
    if roi_gdf is not None:
        # The tiles should intersect, but touching is not enough
        tiles_intersects_roi_gdf = tiles_to_download_gdf.sjoin(
            roi_gdf[["geometry"]], how="inner", predicate="intersects"
        )
        tiles_touches_roi_gdf = tiles_to_download_gdf.sjoin(
            roi_gdf[["geometry"]], how="inner", predicate="touches"
        )
        tiles_to_download_gdf = tiles_intersects_roi_gdf.loc[
            ~tiles_intersects_roi_gdf.index.isin(tiles_touches_roi_gdf.index)
        ]

    # Write the tiles to download to file for reference
    tiles_path = output_image_dir / "tiles_to_download.gpkg"
    gfo.to_file(tiles_to_download_gdf, tiles_path)

    with futures.ThreadPoolExecutor(nb_concurrent_calls) as pool:
        # Loop through all columns and get the images...
        has_switched_axes = _has_switched_axes(crs)
        nb_total = len(tiles_to_download_gdf)
        nb_processed = 0
        nb_downloaded = 0
        download_queue = {}
        logger.info(f"Start loading {nb_total} images")
        progress = None
        for tile in tiles_to_download_gdf.geometry.bounds.itertuples():
            # If a cron_schedule is specified, check if we should be running
            if cron_schedule is not None and cron_schedule != "":
                # Sleep till the schedule becomes active
                first_cron_check = True
                while not pycron.is_now(cron_schedule):
                    # The first time, log message that we are going to sleep...
                    if first_cron_check is True:
                        logger.info(f"Time schedule specified: sleep: {cron_schedule}")
                        first_cron_check = False
                    time.sleep(60)

            # Init progress
            if progress is None:
                message = (
                    f"load_images to {output_image_dir.parent.name}/"
                    f"{output_image_dir.name}"
                )
                progress = progress_util.ProgressLogger(
                    message=message,
                    nb_steps_total=nb_total,
                    nb_steps_done=0,
                )

            nb_processed += 1
            _, tile_xmin, tile_ymin, tile_xmax, tile_ymax = tile
            tile_pixel_width = image_pixel_width
            tile_pixel_height = image_pixel_height

            # If overlapping images are wanted... increase image bbox
            if pixels_overlap:
                tile_xmin -= pixels_overlap * image_crs_pixel_x_size
                tile_xmax += pixels_overlap * image_crs_pixel_x_size
                tile_ymin -= pixels_overlap * image_crs_pixel_y_size
                tile_ymax += pixels_overlap * image_crs_pixel_y_size
                tile_pixel_width += 2 * pixels_overlap
                tile_pixel_height += 2 * pixels_overlap

            # Create output filepath
            if crs.is_projected:
                output_dir = output_image_dir / f"{tile_xmin:06.0f}"
            else:
                output_dir = output_image_dir / f"{tile_xmin:09.4f}"
            output_filename = create_filename(
                crs=crs,
                bbox=(tile_xmin, tile_ymin, tile_xmax, tile_ymax),
                size=(tile_pixel_width, tile_pixel_height),
                image_format=image_format,
                layername=None,
            )
            output_filepath = output_dir / output_filename

            # Do some checks to know if the image needs to be downloaded
            if nb_images_to_skip > 0 and (nb_processed % nb_images_to_skip) != 0:
                # If we need to skip images, do so...
                progress.step()
                continue
            elif (
                not force
                and output_filepath.exists()
                and output_filepath.stat().st_size > 0
            ):
                # Image exists already
                progress.step()
                logger.debug("    -> image exists already, so skip")
                continue

            # Submit the image to be downloaded
            output_dir.mkdir(parents=True, exist_ok=True)
            future = pool.submit(
                getmap_to_file,  # Function
                layersources=layersources,
                output_dir=output_dir,
                crs=crs,
                bbox=(tile_xmin, tile_ymin, tile_xmax, tile_ymax),
                size=(tile_pixel_width, tile_pixel_height),
                ssl_verify=ssl_verify,
                image_format=image_format,
                image_format_save=image_format_save,
                output_filename=output_filename,
                transparent=transparent,
                tiff_compress=tiff_compress,
                image_pixels_ignore_border=image_pixels_ignore_border,
                has_switched_axes=has_switched_axes,
                force=force,
                on_outside_layer_bounds="return",
            )
            download_queue[future] = output_filename

            # Process finished downloads till queue is of acceptable size
            while True:
                # Process downloads that are ready
                futures_done = []
                for future in download_queue:
                    if not future.done():
                        continue

                    # Fetch result: will throw exception if something went wrong
                    _ = future.result()
                    nb_downloaded += 1
                    futures_done.append(future)

                    # Log the progress and download speed
                    progress.step()

                # Remove futures that are done
                for future in futures_done:
                    del download_queue[future]
                futures_done = []

                # If all image tiles have been processed or if the max number of
                # images to download is reached...
                if (
                    nb_processed >= nb_total
                    or max_nb_images > -1
                    and nb_downloaded >= max_nb_images
                ):
                    if len(download_queue) == 0:
                        return
                else:
                    # Not all tiles have been processed yet, and the queue isn't too
                    # full, so process some more
                    if len(download_queue) < nb_concurrent_calls * 2:
                        break

                # Sleep a bit before checking again if there are downloads ready
                time.sleep(0.1)


def align_bbox_to_grid(
    bbox: tuple[float, float, float, float],
    grid_xmin: float,
    grid_ymin: float,
    pixel_size_x: float,
    pixel_size_y: float,
    log_level: int = logging.INFO,
) -> tuple[float, float, float, float]:
    """Align a bounding box to the grid specified.

    Args:
        bbox (tuple[float, float, float, float]): the bounding box
        grid_xmin (float): xmin of the grid to align to.
        grid_ymin (float): ymin of the grid to align to.
        pixel_size_x (float): pixel size for x.
        pixel_size_y (float): pixel size for y.
        log_level (int, optional): the log level to use. Defaults to logging.INFO.

    Returns:
        tuple[float, float, float, float]: the aligned bbox.
    """
    # Make bounds compatible with the grid
    bbox_tmp = list(bbox)
    if (bbox_tmp[0] - grid_xmin) % pixel_size_x != 0:
        xmin_new = bbox_tmp[0] - ((bbox_tmp[0] - grid_xmin) % pixel_size_x)
        msg = f"xmin {bbox_tmp[0]} incompatible with grid, {xmin_new} used"
        logger.log(level=log_level, msg=msg)
        bbox_tmp[0] = xmin_new
    if (bbox_tmp[1] - grid_ymin) % pixel_size_y != 0:
        ymin_new = bbox_tmp[1] - ((bbox_tmp[1] - grid_ymin) % pixel_size_y)
        msg = f"ymin {bbox_tmp[1]} incompatible with grid, {ymin_new} used"
        logger.log(level=log_level, msg=msg)
        bbox_tmp[1] = ymin_new
    if (bbox_tmp[2] - grid_xmin) % pixel_size_x != 0:
        xmax_new = (
            bbox_tmp[2] + pixel_size_x - ((bbox_tmp[2] - grid_xmin) % pixel_size_x)
        )
        msg = f"xmax {bbox_tmp[2]} incompatible with grid, {xmax_new} used"
        logger.log(level=log_level, msg=msg)
        bbox_tmp[2] = xmax_new
    if (bbox_tmp[3] - grid_ymin) % pixel_size_y != 0:
        ymax_new = (
            bbox_tmp[3] + pixel_size_y - ((bbox_tmp[3] - grid_ymin) % pixel_size_y)
        )
        msg = f"ymax {bbox_tmp[3]} incompatible with grid, {ymax_new} used"
        logger.log(level=log_level, msg=msg)
        bbox_tmp[3] = ymax_new

    return (bbox_tmp[0], bbox_tmp[1], bbox_tmp[2], bbox_tmp[3])


def _interprete_ssl_verify(ssl_verify: Union[bool, str, None]):
    # Interprete ssl_verify
    auth = None
    if ssl_verify is not None:
        # If it is a string, make sure it isn't actually a bool
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == "true":
                ssl_verify = True
            elif ssl_verify.lower() == "false":
                ssl_verify = False

        assert isinstance(ssl_verify, bool)
        auth = owslib.util.Authentication(verify=ssl_verify)
        if ssl_verify is False:
            urllib3.disable_warnings()
            logger.warning("SSL VERIFICATION IS TURNED OFF!!!")

    return auth


def getmap_to_file(
    layersources: Union[WMSLayerSource, FileLayerSource, list],
    output_dir: Path,
    crs: Union[str, pyproj.CRS],
    bbox: tuple[float, float, float, float],
    size: tuple[int, int],
    ssl_verify: Union[bool, str] = True,
    image_format: str = FORMAT_GEOTIFF,
    image_format_save: Optional[str] = None,
    output_filename: Optional[str] = None,
    transparent: bool = False,
    tiff_compress: str = "lzw",
    image_pixels_ignore_border: int = 0,
    force: bool = False,
    layername_in_filename: bool = False,
    has_switched_axes: Optional[bool] = None,
    on_outside_layer_bounds: Optional[str] = "raise",
) -> Optional[Path]:
    """Reads/fetches an image from a layer source and saves it to a file.

    Args:
        layersources (WMSLayerSource, FileLayerSource, List): Layer source(s) to get
            images from. Multiple sources can be specified to create a combined image,
            eg. use band 1 of a layersource with band 2 and 3 of another one.
        output_dir (Path): Directory to save the images to.
        crs (pyproj.CRS): The crs of the source and destination images.
        bbox (tuple[float, float, float, float]): Bbox of the image to get.
        size (tuple[int, int]): The image width and height.
        ssl_verify (bool or str, optional): True to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to True.
        image_format (str, optional): [description]. Defaults to FORMAT_GEOTIFF.
        image_format_save (str, optional): [description]. Defaults to None.
        output_filename (str, optional): [description]. Defaults to None.
        transparent (bool, optional): [description]. Defaults to False.
        tiff_compress (str, optional): [description]. Defaults to 'lzw'.
        image_pixels_ignore_border (int, optional): [description]. Defaults to 0.
        force (bool, optional): [description]. Defaults to False.
        layername_in_filename (bool, optional): [description]. Defaults to False.
        has_switched_axes (bool, optional): True if x and y axes should be switched to
            in the WMS GetMap request. If None, an effort is made to determine it
            automatically based on the crs. Defaults to None.
        on_outside_layer_bounds (str, optional): What to do if the bbox asked is outside
            the layer bounds. Defaults to "raise". Options:
            - "raise": raise an error.
            - "return": don't save a file and return None.

    Returns:
        Optional[Path]: The path the file is created at if created succesfully. None if
            the file is not created.
    """
    # Init
    if on_outside_layer_bounds not in ["raise", "return"]:
        raise ValueError(f"Invalid value for {on_outside_layer_bounds=}")

    # If no separate save format is specified, use the standard image_format
    if image_format_save is None:
        image_format_save = image_format

    # Convert input parameters if relevant
    if isinstance(crs, str):
        crs = pyproj.CRS(crs)
    if not isinstance(layersources, list):
        layersources = [layersources]

    # If there isn't a filename supplied, create one...
    if output_filename is None:
        layername = None
        if layername_in_filename:
            for layersource in layersources:
                if layername is None:
                    layername = "_".join(layersource.layernames)
                else:
                    layername += f"_{'_'.join(layersource.layernames)}"

        output_filename = create_filename(
            crs=crs,
            bbox=bbox,
            size=size,
            image_format=image_format_save,
            layername=layername,
        )

    # Create full output filepath
    output_filepath = output_dir / output_filename

    # If force is false and file exists already, stop...
    if force is False and output_filepath.exists():
        if output_filepath.stat().st_size > 0:
            logger.debug(f"File already exists, skip: {output_filepath}")
            return None
        else:
            output_filepath.unlink()

    logger.debug(f"Get image to {output_filepath}")

    if not output_dir.exists():
        output_dir.mkdir()

    # Get image(s), read the band to keep and save
    # Some hacks for special cases...
    bbox_for_getmap = bbox
    size_for_getmap = size
    x_pixsize = (bbox[2] - bbox[0]) / size[0]
    y_pixsize = (bbox[3] - bbox[1]) / size[1]

    # Dirty hack to ask a bigger picture, and then remove the border again!
    if image_pixels_ignore_border > 0:
        bbox_for_getmap = (
            bbox[0] - x_pixsize * image_pixels_ignore_border,
            bbox[1] - y_pixsize * image_pixels_ignore_border,
            bbox[2] + x_pixsize * image_pixels_ignore_border,
            bbox[3] + y_pixsize * image_pixels_ignore_border,
        )
        size_for_getmap = (
            size[0] + 2 * image_pixels_ignore_border,
            size[1] + 2 * image_pixels_ignore_border,
        )

    # For coordinate systems with switched axis (y, x or lon, lat), switch x and y
    if has_switched_axes is None:
        has_switched_axes = _has_switched_axes(crs)
    if has_switched_axes:
        bbox_for_getmap = (
            bbox_for_getmap[1],
            bbox_for_getmap[0],
            bbox_for_getmap[3],
            bbox_for_getmap[2],
        )

    image_data_output = None
    image_profile_output = None
    response = None
    for layersource in layersources:
        window = None
        memfile = None
        image_file = None
        rio_read_kwargs = {}
        try:
            # If it is a WMS layer source
            if isinstance(layersource, WMSLayerSource):
                # Initialize WMS if this hasn't been done yet
                if layersource.wms_service is None:
                    auth = _interprete_ssl_verify(ssl_verify=ssl_verify)
                    wms_service = owslib.wms.WebMapService(
                        url=layersource.wms_server_url,
                        version=layersource.wms_version,
                        username=layersource.username,
                        password=layersource.password,
                        auth=auth,
                    )
                    if layersource.wms_ignore_capabilities_url:
                        # If the wms url in capabilities should be ignored,
                        # overwrite with original url
                        nb = len(wms_service.getOperationByName("GetMap").methods)
                        for method_id in range(nb):
                            wms_service.getOperationByName("GetMap").methods[method_id][
                                "url"
                            ] = layersource.wms_server_url
                    layersource.wms_service = wms_service

                # Get image from server, and retry up to 10 times...
                nb_retries = 0
                time_sleep = 5
                image_retrieved = False
                while image_retrieved is False:
                    try:
                        logger.debug(f"Start call GetMap for bbox {bbox}")
                        response = layersource.wms_service.getmap(
                            layers=layersource.layernames,
                            styles=layersource.layerstyles,
                            srs=f"epsg:{crs.to_epsg()}",
                            bbox=bbox_for_getmap,
                            size=size_for_getmap,
                            format=image_format,
                            transparent=transparent,
                        )
                        logger.debug(f"Finished doing request {response.geturl()}")

                        # If a random sleep was specified... apply it
                        if layersource.random_sleep > 0:
                            time.sleep(random.uniform(0, layersource.random_sleep))

                        # Image was retrieved... so stop loop
                        image_retrieved = True
                    except Exception as ex:
                        if isinstance(ex, owslib.util.ServiceException):
                            if "Error rendering coverage on the fast path" in str(ex):
                                message = f"WMS error for bbox {bbox_for_getmap}: {ex}"
                                if on_outside_layer_bounds == "return":
                                    logger.error(message)
                                    return None
                                else:
                                    raise RuntimeError(message) from ex
                            elif "java.lang.OutOfMemoryError: Java heap" in str(ex):
                                logger.debug(
                                    f"Request for bbox {bbox_for_getmap} gave an "
                                    f"exception, try again in {time_sleep} s: {ex}"
                                )
                            else:
                                message = f"WMS error for bbox {bbox_for_getmap}: {ex}"
                                raise RuntimeError(message) from ex

                        # If the exception isn't handled yet, retry 10 times...
                        if nb_retries < 10:
                            time.sleep(time_sleep)

                            # Increase sleep time every
                            # time.
                            time_sleep += 5
                            nb_retries += 1
                            continue
                        else:
                            message = (
                                "Retried 10 times and didn't work, with "
                                f"layers: {layersource.layernames}, "
                                f"styles: {layersource.layerstyles}, "
                                f"for bbox: {bbox_for_getmap}"
                            )
                            raise RuntimeError(message) from ex

                # If the response is None, error
                if response is None:
                    raise RuntimeError("No valid response retrieved...")

                # Open the response as a file
                memfile = rio.MemoryFile(response.read())
                # Because the image returned by WMS doesn't contain georeferencing
                # info, suppress NotGeoreferencedWarning
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", category=rio_errors.NotGeoreferencedWarning
                    )
                    image_file = memfile.open()

            elif isinstance(layersource, FileLayerSource):
                image_file = rio.open(str(layersource.path))
                if layersource.bands is not None:
                    nb_bands = len(layersource.bands)
                else:
                    nb_bands = image_file.profile["count"]
                window = rio_windows.from_bounds(
                    left=bbox_for_getmap[0],
                    bottom=bbox_for_getmap[1],
                    right=bbox_for_getmap[2],
                    top=bbox_for_getmap[3],
                    transform=image_file.transform,
                )
                rio_read_kwargs = {
                    "window": window,
                    "out_shape": (nb_bands, size[1], size[0]),
                    "resampling": rasterio.enums.Resampling.nearest,
                    "boundless": True,
                }
            else:
                raise ValueError(f"Unsupported layer source: {layersource}")

            # Read the data we need from the opened file
            if layersource.bands is None:
                # If no specific bands specified, read them all...
                if image_data_output is None:
                    image_data_output = image_file.read(**rio_read_kwargs)
                else:
                    image_data_output = np.append(
                        image_data_output,
                        image_file.read(**rio_read_kwargs),
                        axis=0,
                    )
            elif len(layersource.bands) == 1 and layersource.bands[0] == -1:
                # If 1 band, -1 specified: dirty hack to use greyscale
                # version of rgb image
                image_data_tmp = image_file.read(**rio_read_kwargs)
                image_data_grey = np.mean(image_data_tmp, axis=0).astype(
                    image_data_tmp.dtype
                )
                new_shape = (
                    1,
                    image_data_grey.shape[0],
                    image_data_grey.shape[1],
                )
                image_data_grey = np.reshape(image_data_grey, new_shape)
                if image_data_output is None:
                    image_data_output = image_data_grey
                else:
                    image_data_output = np.append(
                        image_data_output, image_data_grey, axis=0
                    )
            else:
                # If bands specified, only read the bands to keep...
                for band in layersource.bands:
                    # Read the band needed + reshape. Remark: rasterio uses
                    # 1-based indexing instead of 0-based
                    rio_read_kwargs["indexes"] = band + 1
                    image_data_curr = image_file.read(**rio_read_kwargs)
                    new_shape = (
                        1,
                        image_data_curr.shape[0],
                        image_data_curr.shape[1],
                    )
                    image_data_curr = np.reshape(image_data_curr, new_shape)

                    # Set or append to image_data_output
                    if image_data_output is None:
                        image_data_output = image_data_curr
                    else:
                        image_data_output = np.append(
                            image_data_output, image_data_curr, axis=0
                        )

            # Set output profile
            # (# of bands will be corrected later if needed)
            if image_profile_output is None:
                image_profile_output = dict(image_file.profile)
                if window is not None:
                    # Not entire file read, so update tranform with window used
                    # transform = image_file.window_transform(window)
                    # Prarameter to create Affine:
                    #     (x_pixsize, 0.0, xmin, 0.0, -y_pixsize, ymax)
                    transform = rio.Affine(
                        x_pixsize,
                        0.0,
                        bbox_for_getmap[0],
                        0.0,
                        -y_pixsize,
                        bbox_for_getmap[3],
                    )
                    image_profile_output["transform"] = transform

                    assert isinstance(image_data_output, np.ndarray)
                    image_profile_output["height"] = image_data_output.shape[1]
                    image_profile_output["width"] = image_data_output.shape[2]

        finally:
            if image_file is not None:
                image_file.close()
                image_file = None
            if memfile is not None:
                memfile.close()
                memfile = None

    # Write (temporary) output file
    assert image_profile_output is not None
    assert isinstance(image_data_output, np.ndarray)

    # Set correct output driver in profile
    image_profile_output["driver"] = _get_driver_for_image_format(image_format_save)

    # Prepare output bands and set them correctly in profile
    if (
        image_format_save in [FORMAT_JPEG, FORMAT_PNG]
        and image_data_output.shape[0] == 2
    ):
        zero_band = np.zeros(
            shape=(1, image_data_output.shape[1], image_data_output.shape[2]),
            dtype=image_data_output.dtype,
        )
        image_data_output = np.append(image_data_output, zero_band, axis=0)

    assert isinstance(image_data_output, np.ndarray)
    image_profile_output["count"] = image_data_output.shape[0]
    image_profile_output = _get_cleaned_write_profile(image_profile_output)

    # Because the (temporary) output file doesn't contain coordinates (yet),
    # suppress NotGeoreferencedWarning while writing
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=rio_errors.NotGeoreferencedWarning)

        try:
            with rio.open(
                str(output_filepath), "w", **image_profile_output
            ) as image_file:
                image_file.write(image_data_output)
        except CPLE_AppDefinedError as ex:  # pragma: no cover
            if ex.errmsg.startswith("Deleting ") and ex.errmsg.endswith(
                " failed: No such file or directory"
            ):
                # Occasionally this error occurs, not sure why: ignore it.
                logger.debug(f"Ignore error: {ex}")
            else:
                raise ex

    # If an aux.xml file was written, remove it again...
    output_aux_path = output_filepath.parent / f"{output_filepath.name}.aux.xml"
    try:
        output_aux_path.unlink(missing_ok=True)
    except Exception as ex:  # pragma: no cover
        # Occasionally the .aux.xml file is locked, not sure why: ignore it.
        logger.debug(f"Ignore error: {ex}")

    # Make the output image compliant with image_format_save

    # If geotiff is asked, check if the the coordinates are embedded...
    if image_format_save == FORMAT_GEOTIFF:
        # Read output image to check if coördinates are there
        with rio.open(str(output_filepath)) as image_file:
            image_profile_orig = image_file.profile
            image_transform_affine = image_file.transform

            if image_pixels_ignore_border == 0:
                image_data_output = image_file.read()
            else:
                image_data_output = image_file.read(
                    window=rio_windows.Window(
                        col_off=image_pixels_ignore_border,
                        row_off=image_pixels_ignore_border,
                        width=size[0],
                        height=size[1],
                    )
                )

        logger.debug(f"original image_profile: {image_profile_orig}")

        # If coordinates are not embedded add them, if image_pixels_ignore_border
        # change them
        if (
            image_transform_affine[2] == 0 and image_transform_affine[5] == 0
        ) or image_pixels_ignore_border > 0:
            logger.debug(
                f"Coordinates not in image, driver: {image_profile_orig['driver']}"
            )

            # If profile format is not gtiff, create new profile
            if image_profile_orig["driver"] != "GTiff":
                image_profile_gtiff = rio_profiles.DefaultGTiffProfile.defaults

                # Copy appropriate info from source file
                image_profile_gtiff.update(
                    count=image_profile_orig["count"],
                    width=size[0],
                    height=size[1],
                    nodata=image_profile_orig["nodata"],
                    dtype=image_profile_orig["dtype"],
                )
                image_profile = image_profile_gtiff
            else:
                image_profile = image_profile_orig

            # Set the asked compression
            image_profile.update(compress=tiff_compress)

            # For some coordinate systems apparently the axis ordered is wrong in LibOWS
            crs_pixel_x_size = (bbox[2] - bbox[0]) / size[0]
            crs_pixel_y_size = (bbox[1] - bbox[3]) / size[1]

            logger.debug(
                "Coordinates to put in geotiff:\n"
                f"    - x-part of pixel width, W-E: {crs_pixel_x_size}\n"
                "    - y-part of pixel width, W-E (0 if image is exactly N up): 0\n"
                f"    - top-left x: {bbox[0]}\n"
                "    - x-part of pixel height, N-S (0 if image is exactly N up): \n"
                f"    - y-part of pixel height, N-S: {crs_pixel_y_size}\n"
                f"    - top-left y: {bbox[3]}"
            )

            # Add transform and crs to the profile
            image_profile.update(
                transform=rio_transform.Affine(
                    crs_pixel_x_size, 0, bbox[0], 0, crs_pixel_y_size, bbox[3]
                ),
                crs=crs,
            )

            # Delete output file, and write again
            output_filepath.unlink()
            with rio.open(str(output_filepath), "w", **image_profile) as image_file:
                image_file.write(image_data_output)

    else:
        # For file formats that doesn't support coordinates, we add a worldfile
        crs_pixel_x_size = (bbox[2] - bbox[0]) / size[0]
        crs_pixel_y_size = (bbox[1] - bbox[3]) / size[1]

        path_noext = output_filepath.parent / output_filepath.stem
        ext_world = _get_world_ext_for_image_format(image_format_save)
        output_worldfile_filepath = Path(str(path_noext) + ext_world)

        with output_worldfile_filepath.open("w") as wld_file:
            wld_file.write(f"{crs_pixel_x_size}")
            wld_file.write("\n0.000")
            wld_file.write("\n0.000")
            wld_file.write(f"\n{crs_pixel_y_size}")
            wld_file.write(f"\n{bbox[0]}")
            wld_file.write(f"\n{bbox[3]}")

        # If the image format to save is different, or if a border needs to be ignored
        if image_format != image_format_save or image_pixels_ignore_border > 0:
            # Read image
            with rio.open(str(output_filepath)) as image_file:
                image_profile_orig = image_file.profile
                image_transform_affine = image_file.transform

                # If border needs to be ignored, only read data we are interested in
                if image_pixels_ignore_border == 0:
                    image_data_output = image_file.read()
                else:
                    image_data_output = image_file.read(
                        window=rio_windows.Window(
                            col_off=image_pixels_ignore_border,
                            row_off=image_pixels_ignore_border,
                            width=size[0],
                            height=size[1],
                        )
                    )

            # If same save format, reuse profile
            if image_format == image_format_save:
                image_profile_curr = image_profile_orig
                if image_pixels_ignore_border != 0:
                    image_profile_curr.update(width=size[0], height=size[1])
            else:
                if image_format_save == FORMAT_TIFF:
                    driver = "GTiff"
                    compress = tiff_compress
                else:
                    raise Exception(
                        f"Unsupported image_format_save: {image_format_save}"
                    )
                image_profile_curr = rio_profiles.Profile(
                    width=size[0],
                    height=size[1],
                    count=image_profile_orig["count"],
                    nodata=image_profile_orig["nodata"],
                    dtype=image_profile_orig["dtype"],
                    compress=compress,
                    driver=driver,
                )

            # Delete output file, and write again
            output_filepath.unlink()
            image_profile_curr = _get_cleaned_write_profile(image_profile_curr)
            with rio.open(
                str(output_filepath), "w", **image_profile_curr
            ) as image_file:
                image_file.write(image_data_output)

    return output_filepath


def create_filename(
    crs: pyproj.CRS, bbox, size, image_format: str, layername: Optional[str] = None
) -> str:
    """Create filename.

    Args:
        crs (pyproj.CRS): crs
        bbox (_type_): bbox
        size (_type_): size
        image_format (str): format
        layername (Optional[str], optional): layername. Defaults to None.

    Returns:
        str: filename
    """
    # Get image extension based on format
    image_ext = _get_ext_for_image_format(image_format)

    # Use different file names for projected vs geographic crs
    if crs.is_projected:
        output_filename = (
            f"{bbox[0]:06.0f}_{bbox[1]:06.0f}_{bbox[2]:06.0f}_{bbox[3]:06.0f}_"
            f"{size[0]}_{size[1]}"
        )
    else:
        output_filename = (
            f"{bbox[0]:09.4f}_{bbox[1]:09.4f}_{bbox[2]:09.4f}_{bbox[3]:09.4f}_"
            f"{size[0]}_{size[1]}"
        )

    # Add layername if it is not None
    if layername is not None:
        output_filename += "_" + layername

    # Add extension
    output_filename += image_ext

    return output_filename


def _has_switched_axes(crs: pyproj.CRS):
    if len(crs.axis_info) < 2:
        logger.warning(
            f"has_switched_axes False: len(crs_31370.axis_info) < 2 for {crs}"
        )

    has_switched_axes_options = [
        {"abbrev": "x", "direction": "east", "has_switched_axes": False},
        {"abbrev": "y", "direction": "north", "has_switched_axes": False},
        {"abbrev": "x", "direction": "north", "has_switched_axes": True},
        {"abbrev": "y", "direction": "east", "has_switched_axes": True},
        {"abbrev": "lat", "direction": "north", "has_switched_axes": False},
        {"abbrev": "lon", "direction": "east", "has_switched_axes": False},
        {"abbrev": "lat", "direction": "east", "has_switched_axes": True},
        {"abbrev": "lon", "direction": "north", "has_switched_axes": True},
    ]
    for axis_info in crs.axis_info:
        for option in has_switched_axes_options:
            if (
                axis_info.abbrev.lower() == option["abbrev"]
                and axis_info.direction.lower() == option["direction"]
            ):
                return option["has_switched_axes"]

    logger.warning(f"has_switched_axes option not found, so assume False for {crs}")
    return False


def _get_driver_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_DRIVER
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_DRIVER
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_DRIVER
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_DRIVER
    else:
        raise Exception(
            f"get_ext_for_image_format for image format {image_format} not implemented"
        )


def _get_ext_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_EXT
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_EXT
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_EXT
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_EXT
    else:
        raise Exception(
            f"get_ext_for_image_format for image format {image_format} not implemented"
        )


def _get_world_ext_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_EXT_WORLD
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_EXT_WORLD
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_EXT_WORLD
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_EXT_WORLD
    else:
        raise Exception(
            f"get_world_ext_for_image_format for format {image_format} not implemented"
        )


def _get_cleaned_write_profile(
    profile: Union[dict, rio_profiles.Profile],
) -> Union[dict, rio_profiles.Profile]:
    # Depending on the driver, different profile keys are supported
    if profile.get("driver") == "JPEG":
        # Don't copy profile keys to cleaned version that are not supported for JPEG
        profile_cleaned = {}
        for profile_key in profile:
            if profile_key not in [
                "blockxsize",
                "blockysize",
                "compress",
                "interleave",
                "photometric",
                "tiled",
            ]:
                profile_cleaned[profile_key] = profile[profile_key]
    elif profile.get("driver") == "PNG":
        # Don't copy profile keys to cleaned version that are not supported for PNG
        profile_cleaned = {}
        for profile_key in profile:
            if profile_key not in [
                "blockxsize",
                "blockysize",
                "compress",
                "interleave",
                "tiled",
            ]:
                profile_cleaned[profile_key] = profile[profile_key]
    else:
        profile_cleaned = profile.copy()

    return profile_cleaned



"""Module to make it easy to start a training session."""

import argparse
import gc
import logging
import os
import shutil
import sys
import traceback
from pathlib import Path

from tensorflow import keras as kr

from orthoseg.helpers import config_helper as conf, email_helper
from orthoseg.lib import predicter, prepare_traindatasets as prep, trainer
from orthoseg.model import model_factory as mf, model_helper as mh
from orthoseg.util import log_util

# Get a logger...
logger = logging.getLogger(__name__)


def _train_args(args) -> argparse.Namespace:
    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    optional.add_argument(
        "config_overrules",
        nargs="*",
        help=(
            "Supply any number of config overrules like this: "
            "<section>.<parameter>=<value>"
        ),
    )

    return parser.parse_args(args)


def train(config_path: Path, config_overrules: list[str] = []):
    """Run a training session for the config specified.

    Args:
        config_path (Path): Path to the config file to use.
        config_overrules (list[str], optional): list of config options that will
            overrule other ways to supply configuration. They should be specified in the
            form of "<section>.<parameter>=<value>". Defaults to [].
    """
    # Init
    # Load the config and save in a bunch of global variables zo it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path, overrules=config_overrules)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging_conf.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)

    # Log start
    logger.info(f"Start train for config {config_path.stem}")
    logger.debug(f"Config used: \n{conf.pformat_config()}")

    try:
        # Create the output dir's if they don't exist yet...
        for dir in [
            conf.dirs.getpath("project_dir"),
            conf.dirs.getpath("training_dir"),
        ]:
            if dir and not dir.exists():
                dir.mkdir()

        train_label_infos = conf.get_train_label_infos()

        # Determine the projection of (the first) train layer... it will be used for all
        train_image_layer = train_label_infos[0].image_layer
        train_projection = conf.image_layers[train_image_layer]["projection"]

        classes = conf.determine_classes()

        # Now create the train datasets (train, validation, test)
        force_model_traindata_id = conf.train.getint("force_model_traindata_id")
        if force_model_traindata_id > -1:
            training_dir = (
                conf.dirs.getpath("training_dir") / f"{force_model_traindata_id:02d}"
            )
            traindata_id = force_model_traindata_id
        else:
            logger.info("Prepare train, validation and test data")
            training_dir, traindata_id = prep.prepare_traindatasets(
                label_infos=train_label_infos,
                classes=classes,
                image_layers=conf.image_layers,
                training_dir=conf.dirs.getpath("training_dir"),
                labelname_column=conf.train.get("labelname_column"),
                image_pixel_x_size=conf.train.getfloat("image_pixel_x_size"),
                image_pixel_y_size=conf.train.getfloat("image_pixel_y_size"),
                image_pixel_width=conf.train.getint("image_pixel_width"),
                image_pixel_height=conf.train.getint("image_pixel_height"),
                ssl_verify=conf.general["ssl_verify"],
            )

        # Send mail that we are starting train
        email_helper.sendmail(f"Start train for config {config_path.stem}")
        logger.info(
            f"Traindata dir to use is {training_dir}, with traindata_id: {traindata_id}"
        )
        traindata_dir = training_dir / "train"
        validationdata_dir = training_dir / "validation"
        testdata_dir = training_dir / "test"

        # Check if training is needed
        # Get hyper parameters from the config
        # TODO: activation_function should probably not be specified!!!!!!
        architectureparams = mh.ArchitectureParams(
            architecture=conf.model["architecture"],
            classes=list(classes),
            nb_channels=conf.model.getint("nb_channels"),
            architecture_id=conf.model.getint("architecture_id"),
            activation_function="softmax",
        )
        trainparams = mh.TrainParams(
            trainparams_id=conf.train.getint("trainparams_id"),
            image_augmentations=conf.train.getdict("image_augmentations"),
            mask_augmentations=conf.train.getdict("mask_augmentations"),
            class_weights=[classes[classname]["weight"] for classname in classes],
            batch_size=conf.train.getint("batch_size_fit"),
            optimizer=conf.train.get("optimizer"),
            optimizer_params=conf.train.getdict("optimizer_params"),
            loss_function=conf.train.get("loss_function"),
            monitor_metric=conf.train.get("monitor_metric"),
            monitor_metric_mode=conf.train.get("monitor_metric_mode"),
            save_format=conf.train.get("save_format"),
            save_best_only=conf.train.getboolean("save_best_only"),
            save_min_accuracy=conf.train.getfloat("save_min_accuracy"),
            nb_epoch=conf.train.getint("max_epoch"),
            nb_epoch_with_freeze=conf.train.getint("nb_epoch_with_freeze"),
            earlystop_patience=conf.train.getint("earlystop_patience"),
            earlystop_monitor_metric=conf.train.get("earlystop_monitor_metric"),
            earlystop_monitor_metric_mode=conf.train.get(
                "earlystop_monitor_metric_mode"
            ),
            log_tensorboard=conf.train.getboolean("log_tensorboard"),
            log_csv=conf.train.getboolean("log_csv"),
        )

        # Check if there exists already a model for this train dataset + hyperparameters
        model_dir = conf.dirs.getpath("model_dir")
        segment_subject = conf.general["segment_subject"]
        best_model_curr_train_version = mh.get_best_model(
            model_dir=model_dir,
            segment_subject=segment_subject,
            traindata_id=traindata_id,
            architecture_id=architectureparams.architecture_id,
            trainparams_id=trainparams.trainparams_id,
        )

        # Determine if training is needed,...
        resume_train = conf.train.getboolean("resume_train")
        if resume_train is False:
            # If no (best) model found, training needed!
            if best_model_curr_train_version is None:
                train_needed = True
            elif conf.train.getboolean("force_train") is True:
                train_needed = True
            else:
                logger.info(
                    "JUST PREDICT, no training: resume_train is false + model found"
                )
                train_needed = False
        else:
            # We want to preload an existing model and models were found
            if best_model_curr_train_version is not None:
                logger.info(
                    "PRELOAD model + continue TRAINING: "
                    f"{best_model_curr_train_version['filename']}"
                )
                train_needed = True
            else:
                message = "STOP: preload_existing_model is true but no model was found!"
                logger.error(message)
                raise Exception(message)

        # Train!!!
        min_probability = conf.predict.getfloat("min_probability")
        if train_needed is True:
            # If a model already exists, use it to predict (possibly new) training and
            # validation dataset. This way it is possible to have a quick check on
            # errors in (new) added labels in the datasets.

            # Get the current best model that already exists for this subject
            best_recent_model = mh.get_best_model(
                model_dir=model_dir,
                segment_subject=segment_subject,
                architecture_id=architectureparams.architecture_id,
                trainparams_id=trainparams.trainparams_id,
            )
            if best_recent_model is not None:
                try:
                    # TODO: move the hyperparams filename formatting to get_models...
                    logger.info(
                        f"Load model + weights from {best_recent_model['filepath']}"
                    )
                    best_model = mf.load_model(
                        best_recent_model["filepath"], compile=False
                    )
                    best_hyperparams_path = (
                        best_recent_model["filepath"].parent
                        / f"{best_recent_model['basefilename']}_hyperparams.json"
                    )
                    best_hyperparams = mh.HyperParams(path=best_hyperparams_path)
                    logger.info("Loaded model, weights and params")

                    # Prepare output subdir to be used for predictions
                    predict_out_subdir, _ = os.path.splitext(
                        best_recent_model["filename"]
                    )

                    # Predict training dataset
                    predicter.predict_dir(
                        model=best_model,
                        input_image_dir=traindata_dir / "image",
                        output_image_dir=traindata_dir / predict_out_subdir,
                        output_vector_path=None,
                        projection_if_missing=train_projection,
                        input_mask_dir=traindata_dir / "mask",
                        batch_size=conf.train.getint("batch_size_predict"),
                        evaluate_mode=True,
                        classes=best_hyperparams.architecture.classes,
                        min_probability=min_probability,
                        cancel_filepath=conf.files.getpath("cancel_filepath"),
                        max_prediction_errors=conf.predict.getint(
                            "max_prediction_errors"
                        ),
                    )

                    # Predict validation dataset
                    predicter.predict_dir(
                        model=best_model,
                        input_image_dir=validationdata_dir / "image",
                        output_image_dir=validationdata_dir / predict_out_subdir,
                        output_vector_path=None,
                        projection_if_missing=train_projection,
                        input_mask_dir=validationdata_dir / "mask",
                        batch_size=conf.train.getint("batch_size_predict"),
                        evaluate_mode=True,
                        classes=best_hyperparams.architecture.classes,
                        min_probability=min_probability,
                        cancel_filepath=conf.files.getpath("cancel_filepath"),
                        max_prediction_errors=conf.predict.getint(
                            "max_prediction_errors"
                        ),
                    )
                    del best_model
                except Exception as ex:
                    logger.warning(f"Exception trying to predict with old model: {ex}")

            # Now we can really start training
            logger.info("Start training")
            model_preload_filepath = None
            if best_model_curr_train_version is not None:
                model_preload_filepath = best_model_curr_train_version["filepath"]
            elif conf.train.getboolean("preload_with_previous_traindata"):
                best_model_for_architecture = mh.get_best_model(
                    model_dir=model_dir, segment_subject=segment_subject
                )
                if best_model_for_architecture is not None:
                    model_preload_filepath = best_model_for_architecture["filepath"]

            # Combine all hyperparameters in hyperparams object
            hyperparams = mh.HyperParams(
                architecture=architectureparams, train=trainparams
            )

            trainer.train(
                traindata_dir=traindata_dir,
                validationdata_dir=validationdata_dir,
                model_save_dir=model_dir,
                segment_subject=segment_subject,
                traindata_id=traindata_id,
                hyperparams=hyperparams,
                model_preload_filepath=model_preload_filepath,
                image_width=conf.train.getint("image_pixel_width"),
                image_height=conf.train.getint("image_pixel_height"),
                save_augmented_subdir=conf.train.get("save_augmented_subdir"),
            )

            # Now get the best model found during training
            best_model_curr_train_version = mh.get_best_model(
                model_dir=model_dir,
                segment_subject=segment_subject,
                traindata_id=traindata_id,
            )

        # Assert to evade typing warnings
        assert best_model_curr_train_version is not None

        # Now predict on the train,... data
        logger.info(
            "PREDICT test data with best model: "
            f"{best_model_curr_train_version['filename']}"
        )

        # Load prediction model...
        logger.info(
            f"Load model + weights from {best_model_curr_train_version['filepath']}"
        )
        model = mf.load_model(best_model_curr_train_version["filepath"], compile=False)
        logger.info("Loaded model + weights")

        # Prepare output subdir to be used for predictions
        predict_out_subdir, _ = os.path.splitext(
            best_model_curr_train_version["filename"]
        )

        # Predict training dataset
        predicter.predict_dir(
            model=model,
            input_image_dir=traindata_dir / "image",
            output_image_dir=traindata_dir / predict_out_subdir,
            output_vector_path=None,
            projection_if_missing=train_projection,
            input_mask_dir=traindata_dir / "mask",
            batch_size=conf.train.getint("batch_size_predict"),
            evaluate_mode=True,
            classes=classes,
            min_probability=min_probability,
            cancel_filepath=conf.files.getpath("cancel_filepath"),
            max_prediction_errors=conf.predict.getint("max_prediction_errors"),
        )

        # Predict validation dataset
        predicter.predict_dir(
            model=model,
            input_image_dir=validationdata_dir / "image",
            output_image_dir=validationdata_dir / predict_out_subdir,
            output_vector_path=None,
            projection_if_missing=train_projection,
            input_mask_dir=validationdata_dir / "mask",
            batch_size=conf.train.getint("batch_size_predict"),
            evaluate_mode=True,
            classes=classes,
            min_probability=min_probability,
            cancel_filepath=conf.files.getpath("cancel_filepath"),
            max_prediction_errors=conf.predict.getint("max_prediction_errors"),
        )

        # Predict test dataset, if it exists
        if testdata_dir is not None and testdata_dir.exists():
            predicter.predict_dir(
                model=model,
                input_image_dir=testdata_dir / "image",
                output_image_dir=testdata_dir / predict_out_subdir,
                output_vector_path=None,
                projection_if_missing=train_projection,
                input_mask_dir=testdata_dir / "mask",
                batch_size=conf.train.getint("batch_size_predict"),
                evaluate_mode=True,
                classes=classes,
                min_probability=min_probability,
                cancel_filepath=conf.files.getpath("cancel_filepath"),
                max_prediction_errors=conf.predict.getint("max_prediction_errors"),
                no_images_ok=True,
            )

        # Predict extra test dataset with random images in the roi, to add to
        # train and/or validation dataset if inaccuracies are found
        # -> this is very useful to find false positives to improve the datasets
        if conf.dirs.getpath("predictsample_image_input_dir").exists():
            predicter.predict_dir(
                model=model,
                input_image_dir=conf.dirs.getpath("predictsample_image_input_dir"),
                output_image_dir=conf.dirs.getpath("predictsample_image_output_basedir")
                / predict_out_subdir,
                output_vector_path=None,
                projection_if_missing=train_projection,
                batch_size=conf.train.getint("batch_size_predict"),
                evaluate_mode=True,
                classes=classes,
                min_probability=min_probability,
                cancel_filepath=conf.files.getpath("cancel_filepath"),
                max_prediction_errors=conf.predict.getint("max_prediction_errors"),
            )

        # Free resources...
        logger.debug("Free resources")
        if model is not None:
            del model
        kr.backend.clear_session()
        gc.collect()

        # Log and send mail
        message = f"Completed train for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running train for task {config_path.stem}"
        logger.exception(message)
        if isinstance(ex, prep.ValidationError):
            message_body = f"Validation error: {ex.to_html()}"
        else:
            message_body = f"Exception: {ex}<br/><br/>{traceback.format_exc()}"
        email_helper.sendmail(subject=message, body=message_body)
        raise Exception(message) from ex
    finally:
        if conf.tmp_dir is not None:
            shutil.rmtree(conf.tmp_dir, ignore_errors=True)


def main():
    """Run train."""
    try:
        # Interprete arguments
        args = _train_args(sys.argv[1:])

        # Run!
        train(config_path=Path(args.config), config_overrules=args.config_overrules)
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



"""Modile with generic Utility functions for vector manipulations."""

import logging
from typing import Optional

import geopandas as gpd
import pandas as pd
import shapely

logger = logging.getLogger(__name__)


def is_onborder(
    gdf: gpd.GeoDataFrame,
    border_bounds: tuple[float, float, float, float],
    onborder_column_name: str = "onborder",
) -> gpd.GeoDataFrame:
    """Add/update the is_onborder column to the GeoDataFrame.

    The column containes these values:
        * 0 if the polygon isn't on the border and
        * 1 if it is.

    Args:
        gdf: input GeoDataFrame
        border_bounds: the bounds (tupple with (xmin, ymin, xmax, ymax)
                       to check against to determine onborder
        onborder_column_name: the column name of the onborder column
    """
    if gdf is None or len(gdf.index) == 0:
        return gdf

    result_gdf = gdf.copy()
    result_gdf[onborder_column_name] = 0

    # Check
    for i, geom_row in result_gdf.iterrows():
        if geom_row["geometry"] is not None and not geom_row["geometry"].is_empty:
            # Check if the geom is on the border of the tile
            geom_bounds = geom_row["geometry"].bounds
            if (
                geom_bounds[0] <= border_bounds[0]
                or geom_bounds[1] <= border_bounds[1]
                or geom_bounds[2] >= border_bounds[2]
                or geom_bounds[3] >= border_bounds[3]
            ):
                result_gdf.loc[i, onborder_column_name] = 1

    assert isinstance(result_gdf, gpd.GeoDataFrame)
    return result_gdf


def is_valid_reason(geoseries: gpd.GeoSeries) -> pd.Series:
    """Get the reason for invalidity of all geometries in the GeoSeries.

    Args:
        geoseries (gpd.GeoSeries): the GeoSeries to check.

    Returns:
        pd.Series: _description_
    """
    # Get result and keep geoseries indexes
    return pd.Series(
        data=shapely.is_valid_reason(geoseries),
        index=geoseries.index,
    )


def reclassify_neighbours(
    gdf: gpd.GeoDataFrame,
    reclassify_column: str,
    query: str,
    border_bounds: Optional[tuple[float, float, float, float]],
    class_background: str = "background",
) -> gpd.GeoDataFrame:
    """Reclassify features to the class of neighbouring features.

    For features that comply to the query, if they have a neighbour (touch/overlap),
    change their classname to that of the neighbour with the longest intersection and
    dissolve to merge them.

    The query should follow the syntax of pandas.query queries and can us the following
    fields:
        - onborder (0 or 1): does the feature touch/overlap the border_bounds. If this
          field is used in the query, the border_bounds parameter is mandatory.
        - area (float): the area of the geometry.
        - perimeter (float): the perimeter of the geometry.

    Args:
        gdf (gpd.GeoDataFrame): input
        reclassify_column (str): column to reclassify.
        query (str): th query to find the features to reclassify.
        border_bounds (Optional[tuple[float, float, float, float]]): the bounds of the
            border to use for th onborder field in the query.
        class_background (str, optional): the classname to treat as background.
            Defaults to "background".
        aggfunc ()

    Raises:
        ValueError: raised if incompatible parameters are passed.

    Returns:
        gpd.GeoDataFrame: the reclassified result
    """
    # Init column info
    columns_orig = list(gdf.columns)

    # Init query + needed data
    query = query.replace("\n", " ")

    def _add_needed_columns(
        inner_gdf: gpd.GeoDataFrame,
        inner_query: str,
        inner_border_bounds: Optional[tuple[float, float, float, float]],
    ) -> gpd.GeoDataFrame:
        gdf_result = inner_gdf.copy()

        # Always add area to know which neighbour is largest
        gdf_result["area"] = gdf_result.geometry.area
        if "perimeter" in inner_query:
            gdf_result["perimeter"] = gdf_result.geometry.length
        if "onborder" in inner_query:
            if inner_border_bounds is None:
                raise ValueError(
                    "query contains onborder, but border_bounds parameter is None"
                )
            gdf_result = is_onborder(gdf_result, inner_border_bounds)

        # Return result
        assert isinstance(gdf_result, gpd.GeoDataFrame)
        return gdf_result

    result_gdf = _add_needed_columns(gdf, query, border_bounds)

    # First remove background polygons that don't match the reclassify query
    nobackground_query = (
        f"{reclassify_column} != '{class_background}' or "
        f"({reclassify_column} == '{class_background}' and "
        f"({query}))"
    )
    # Use copy() to avoid view-versus-copy warnings
    result_gdf = result_gdf.query(nobackground_query).copy()

    # Keep looking for polygons that comply with query and give the same class
    # as neighbour till no changes can be made anymore.
    # Stop after 5 iterations to be sure never to end up in endless loop
    reclassify_max = 5
    result_gdf["no_neighbours"] = 0
    query = f"no_neighbours == 0 and ({query})"
    for reclassify_counter in range(reclassify_max):
        # Loop till no features were changed anymore
        if reclassify_counter > 0:
            assert isinstance(result_gdf, gpd.GeoDataFrame)
            result_gdf = _add_needed_columns(result_gdf, query, border_bounds)

        # Use copy() to avoid view-versus-copy warnings
        result_reclass_gdf = result_gdf.query(query).copy()
        if len(result_reclass_gdf) == 0:
            break
        # Order by area to treat the features found from small to large
        result_reclass_gdf = result_reclass_gdf.sort_values(by=["area"])

        for row in result_reclass_gdf.itertuples():
            # Find neighbours (query returns iloc's, not indexes!)
            neighbours_ilocs = result_gdf.geometry.sindex.query(
                row.geometry, predicate="intersects"
            ).tolist()
            if len(neighbours_ilocs) <= 1:
                result_gdf.loc[[row.Index], ["no_neighbours"]] = 1
                continue

            # Remove yourself
            row_loc = result_gdf.index.get_loc(row.Index)
            neighbours_ilocs.remove(row_loc)

            # Find neighbour with longest intersection
            neighbours_gdf = result_gdf.iloc[neighbours_ilocs]
            inters_geoseries = neighbours_gdf.geometry.intersection(row.geometry)
            max_length_index = inters_geoseries.length.idxmax()

            # Change the class of the smallest one to the oher's class
            class_curr = result_gdf.at[row.Index, reclassify_column]
            class_neighbour = result_gdf.at[max_length_index, reclassify_column]
            if class_curr != class_neighbour:
                # If the neighbour is not a reclass feature or if its area is larger
                # than the current feature, use its class
                if (
                    max_length_index not in result_reclass_gdf.index
                    or result_gdf.at[max_length_index, "area"] >= row.area
                ):
                    result_gdf.loc[[row.Index], [reclassify_column]] = class_neighbour
                else:
                    result_gdf.loc[[max_length_index], [reclassify_column]] = class_curr

        # Remove temp columns + dissolve
        result_gdf = result_gdf[columns_orig + ["no_neighbours"]]
        dissolve_columns = [reclassify_column, "no_neighbours"]

        # If there are extra columns, use aggfunc join to concatenate values
        assert isinstance(result_gdf, gpd.GeoDataFrame)
        if len(columns_orig) > 2:
            result_gdf = result_gdf.dissolve(
                by=dissolve_columns, as_index=False, aggfunc=", ".join
            )
        else:
            result_gdf = result_gdf.dissolve(by=dissolve_columns, as_index=False)

        result_gdf = result_gdf.explode(ignore_index=True)

    # Finalize + make sure there is no background in the output
    result_gdf = result_gdf[columns_orig]
    # Use copy() to avoid view-versus-copy warnings
    result_gdf = result_gdf.query(f"{reclassify_column} != '{class_background}'").copy()
    assert isinstance(result_gdf, gpd.GeoDataFrame)
    return result_gdf



"""Script to load images from a WMS server."""

import argparse
import logging
import shlex
import shutil
import sys
import traceback
from pathlib import Path

import pyproj

import orthoseg.model.model_factory as mf
from orthoseg.helpers import config_helper as conf, email_helper
from orthoseg.util import log_util, ows_util

# Get a logger...
logger = logging.getLogger(__name__)


def _load_images_argstr(argstr):
    args = shlex.split(argstr)
    _load_images_args(args)


def _load_images_args(args):
    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    optional.add_argument(
        "config_overrules",
        nargs="*",
        help=(
            "Supply any number of config overrules like this: "
            "<section>.<parameter>=<value>"
        ),
    )

    # Interprete arguments
    args = parser.parse_args(args)

    # Run!
    load_images(config_path=Path(args.config), config_overrules=args.config_overrules)


def load_images(
    config_path: Path,
    load_testsample_images: bool = False,
    config_overrules: list[str] = [],
):
    """Load and cache images for a segmentation project.

    Args:
        config_path (Path): Path to the projects config file.
        load_testsample_images (bool, optional): True to only load testsample
            images. Defaults to False.
        config_overrules (list[str], optional): list of config options that will
            overrule other ways to supply configuration. They should be specified in the
            form of "<section>.<parameter>=<value>". Defaults to [].
    """
    # Init
    # Load the config and save in a bunch of global variables zo it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path, overrules=config_overrules)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging_conf.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)

    # Log + send email
    message = f"Start load_images for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")
    email_helper.sendmail(message)

    try:
        # Use different setting depending if testsample or all images
        if load_testsample_images:
            output_image_dir = conf.dirs.getpath("predictsample_image_input_dir")

            # Use the same image size as for the training, that is the most
            # convenient to check the quality
            image_pixel_width = conf.train.getint("image_pixel_width")
            image_pixel_height = conf.train.getint("image_pixel_height")
            image_pixel_x_size = conf.train.getfloat("image_pixel_x_size")
            image_pixel_y_size = conf.train.getfloat("image_pixel_y_size")
            image_pixels_overlap = 0
            image_format = ows_util.FORMAT_JPEG

            # To create the testsample, fetch only on every ... images
            nb_images_to_skip = 50

        else:
            output_image_dir = conf.dirs.getpath("predict_image_input_dir")

            # Get the image size for the predict
            image_pixel_width = conf.predict.getint("image_pixel_width")
            image_pixel_height = conf.predict.getint("image_pixel_height")
            image_pixel_x_size = conf.predict.getfloat("image_pixel_x_size")
            image_pixel_y_size = conf.predict.getfloat("image_pixel_y_size")
            image_pixels_overlap = conf.predict.getint("image_pixels_overlap", 0)
            image_format = ows_util.FORMAT_JPEG

            # For the real prediction dataset, no skipping obviously...
            nb_images_to_skip = 0

        # Validate the image size for the model architecture
        input_width_pred = image_pixel_width + 2 * image_pixels_overlap
        input_height_pred = image_pixel_height + 2 * image_pixels_overlap
        mf.check_image_size(
            architecture=conf.model.get("architecture"),
            input_width=input_width_pred,
            input_height=input_height_pred,
        )

        # Get ssl_verify setting
        ssl_verify = conf.general["ssl_verify"]
        # Get the download cron schedule
        download_cron_schedule = conf.download["cron_schedule"]

        # Get the layer info
        predict_layer = conf.predict["image_layer"]
        layersources = conf.image_layers[predict_layer]["layersources"]
        nb_concurrent_calls = conf.image_layers[predict_layer]["nb_concurrent_calls"]
        crs = pyproj.CRS.from_user_input(conf.image_layers[predict_layer]["projection"])
        bbox = conf.image_layers[predict_layer]["bbox"]
        grid_xmin = conf.image_layers[predict_layer]["grid_xmin"]
        grid_ymin = conf.image_layers[predict_layer]["grid_ymin"]
        image_pixels_ignore_border = conf.image_layers[predict_layer][
            "image_pixels_ignore_border"
        ]
        roi_filepath = conf.image_layers[predict_layer]["roi_filepath"]

        # Now we are ready to get the images...
        ows_util.get_images_for_grid(
            layersources=layersources,
            output_image_dir=output_image_dir,
            crs=crs,
            image_gen_bbox=bbox,
            image_gen_roi_filepath=roi_filepath,
            grid_xmin=grid_xmin,
            grid_ymin=grid_ymin,
            image_crs_pixel_x_size=image_pixel_x_size,
            image_crs_pixel_y_size=image_pixel_y_size,
            image_pixel_width=image_pixel_width,
            image_pixel_height=image_pixel_height,
            image_pixels_ignore_border=image_pixels_ignore_border,
            nb_concurrent_calls=nb_concurrent_calls,
            cron_schedule=download_cron_schedule,
            image_format=image_format,
            pixels_overlap=image_pixels_overlap,
            nb_images_to_skip=nb_images_to_skip,
            ssl_verify=ssl_verify,
        )

        # Log and send mail
        message = f"Completed load_images for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running load_images for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(
            subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}"
        )
        raise Exception(message) from ex
    finally:
        if conf.tmp_dir is not None:
            shutil.rmtree(conf.tmp_dir, ignore_errors=True)


def main():
    """Run load images."""
    try:
        _load_images_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



"""Module with specific helper functions to manage the configuration of orthoseg."""

import configparser
import json
import logging
import pprint
import re
import tempfile
from pathlib import Path
from typing import Any, Optional, Union

from orthoseg.lib.prepare_traindatasets import LabelInfo
from orthoseg.util import config_util
from orthoseg.util.ows_util import FileLayerSource, WMSLayerSource

# Get a logger...
logger = logging.getLogger(__name__)

# Define the chars that cannot be used in codes that are use in filenames.
# Remark: '_' cannot be used because '_' is used as devider to parse filenames, and if
# it is used in codes as well the parsing becomes a lot more difficult.
illegal_chars_in_codes = ["_", ",", ".", "?", ":"]

tmp_dir: Optional[Path] = None

config: configparser.ConfigParser
config_paths: list[Path]
config_filepaths_used: list[Path]
config_overrules: Any
config_overrules_path: Optional[Path]
general: Any
model: Any
download: Any
train: Any
predict: Any
postprocess: Any
dirs: Any
files: Any
email: Any
logging_conf: Any
cleanup: Any

layer_config_filepath_used: Path
image_layers: dict


def pformat_config() -> str:
    """Format the config to a string.

    Returns:
        str: the configuration as string
    """
    message = f"Config files used: {pprint.pformat(config_filepaths_used)} \n"
    message += f"Layer config file used: {layer_config_filepath_used} \n"
    message += "Config info listing:\n"
    message += pprint.pformat(config_util.as_dict(config))
    message += "Layer config info listing:\n"
    message += pprint.pformat(image_layers)
    return message


def read_orthoseg_config(config_path: Path, overrules: list[str] = []):
    """Read an orthoseg configuration file.

    Args:
        config_path (Path): path to the configuration file to read.
        overrules (list[str], optional): list of config options that will overrule other
            ways to supply configuration. They should be specified as a list of
            "<section>.<parameter>=<value>" strings. Defaults to [].
    """
    # Determine list of config files that should be loaded
    config_paths = config_util.get_config_files(config_path)

    # If there are overrules, write them to a temporary configuration file.
    global config_overrules
    config_overrules = overrules
    global config_overrules_path
    config_overrules_path = None
    if len(config_overrules) > 0:
        config_overrules_path = get_tmp_dir() / "config_overrules.ini"

        # Create config parser, add all overrules
        overrules_parser = configparser.ConfigParser()
        for overrule in config_overrules:
            parts = overrule.split("=")
            if len(parts) != 2:
                raise ValueError(f"invalid config overrule found: {overrule}")
            key, value = parts
            parts2 = key.split(".")
            if len(parts2) != 2:
                raise ValueError(f"invalid config overrule found: {overrule}")
            section, parameter = parts2
            if section not in overrules_parser:
                overrules_parser[section] = {}
            overrules_parser[section][parameter] = value

        # Write to temp file and add file to config_paths
        with open(config_overrules_path, "w") as overrules_file:
            overrules_parser.write(overrules_file)
        config_paths.append(config_overrules_path)

    # Load configs
    global config
    config = config_util.read_config_ext(config_paths)

    # Now do orthoseg-specific checks, inits,... on config
    global config_filepaths_used
    config_filepaths_used = config_paths

    # Now set global variables to each section as shortcuts
    global general
    general = config["general"]
    global model
    model = config["model"]
    global download
    download = config["download"]
    global train
    train = config["train"]
    global predict
    predict = config["predict"]
    global postprocess
    postprocess = config["postprocess"]
    global dirs
    dirs = config["dirs"]
    global files
    files = config["files"]
    global logging_conf
    logging_conf = config["logging"]
    global email
    email = config["email"]
    global cleanup
    cleanup = config["cleanup"]

    # Some checks to make sure the config is loaded properly
    segment_subject = general.get("segment_subject")
    if segment_subject is None or segment_subject == "MUST_OVERRIDE":
        raise ValueError(
            "Projectconfig parameter general.segment_subject needs to be overruled to"
            "a proper name in a specific project config file, \nwith config_filepaths "
            f"{config_paths}"
        )
    elif any(
        illegal_character in segment_subject
        for illegal_character in illegal_chars_in_codes
    ):
        raise ValueError(
            f"Projectconfig parameter general.segment_subject ({segment_subject}) "
            f"should not contain any of the following chars: {illegal_chars_in_codes}"
        )

    # If the projects_dir parameter is a relative path, resolve it towards the location
    # of the project config file.
    projects_dir = dirs.getpath("projects_dir")
    if not projects_dir.is_absolute():
        projects_dir_absolute = (config_path.parent / projects_dir).resolve()
        logger.info(
            f"dirs.projects_dir was relative: is resolved to {projects_dir_absolute}"
        )
        dirs["projects_dir"] = projects_dir_absolute.as_posix()

    # Read the layer config
    layer_config_filepath = files.getpath("image_layers_config_filepath")
    global layer_config_filepath_used
    layer_config_filepath_used = layer_config_filepath

    global image_layers
    image_layers = _read_layer_config(layer_config_filepath=layer_config_filepath)


def get_tmp_dir() -> Path:
    """Get a temporary directory for this run.

    If no temporary directory exists yet, it is created.

    Returns:
        Path: the path to the temporary directory.
    """
    global tmp_dir

    if tmp_dir is None:
        tmp_dir = Path(tempfile.gettempdir()) / "orthoseg"
        tmp_dir.mkdir(parents=True, exist_ok=True)
        tmp_dir = Path(tempfile.mkdtemp(prefix="run_", dir=tmp_dir))

    return tmp_dir


def get_train_label_infos() -> list[LabelInfo]:
    """Searches and returns LabelInfos that can be used to create a training dataset.

    Returns:
        list[LabelInfo]: List of LabelInfos found.
    """
    train_label_infos = _prepare_train_label_infos(
        labelpolygons_pattern=train.getpath("labelpolygons_pattern"),
        labellocations_pattern=train.getpath("labellocations_pattern"),
        label_datasources=train.getdict("label_datasources", None),
        image_layers=image_layers,
    )
    if train_label_infos is None or len(train_label_infos) == 0:
        raise ValueError(
            "No valid label file config found in train.label_datasources or "
            f"with patterns {train.get('labelpolygons_pattern')} and "
            f"{train.get('labellocations_pattern')}"
        )
    return train_label_infos


def determine_classes():
    """Determine classes.

    Raises:
        Exception: Error reading classes

    Returns:
        any: classes
    """
    try:
        classes = train.getdict("classes")

        # If the burn_value property isn't supplied for the classes, add them
        for class_id, (classname) in enumerate(classes):
            if "burn_value" not in classes[classname]:
                classes[classname]["burn_value"] = class_id
        return classes
    except Exception as ex:
        raise Exception(f"Error reading classes: {train.get('classes')}") from ex


def _read_layer_config(layer_config_filepath: Path) -> dict:
    # Init
    if not layer_config_filepath.exists():
        raise Exception(f"Layer config file not found: {layer_config_filepath}")

    # Read config file...
    layer_config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation(),
        converters={
            "list": lambda x: _str2list(x),
            "listint": lambda x: _str2intlist(x),
            "dict": lambda x: None if x is None else json.loads(x),
            "path": lambda x: Path(x),
        },
    )
    layer_config.read(layer_config_filepath)

    # Prepare data
    image_layers: dict[str, dict] = {}
    for image_layer in layer_config.sections():
        # First check if the image_layer code doesn't contain 'illegal' characters
        if any(illegal_char in image_layer for illegal_char in illegal_chars_in_codes):
            raise ValueError(
                f"Section name [{image_layer}] in layer config should not contain any"
                f"of these chars: {illegal_chars_in_codes}, in {layer_config_filepath}"
            )

        # Init layer with all parameters in the section as dict
        image_layers[image_layer] = dict(layer_config[image_layer])

        # Check if the mandatory layer-level properties are present
        if "projection" not in image_layers[image_layer]:
            raise ValueError(
                f"Image layer [{image_layer}] in layer config needs a 'projection' key!"
            )

        # If the layer source(s) are specified in a json parameter, parse it
        if "layersources" in image_layers[image_layer]:
            image_layers[image_layer]["layersources"] = layer_config[
                image_layer
            ].getdict("layersources")
        else:
            # If not, the layersource is specified in some top-level parameters
            layersource = {}
            layersource_keys = [
                "wms_server_url",
                "wms_version",
                "wms_layernames",
                "wms_layerstyles",
                "bands",
                "wms_username",
                "wms_password",
                "random_sleep",
                "wms_ignore_capabilities_url",
                "path",
                "layername",
            ]
            for key in layersource_keys:
                if key in layer_config[image_layer]:
                    layersource[key] = layer_config[image_layer][key]
            image_layers[image_layer]["layersources"] = [layersource]

        # Convert the layersource dicts to layersource objects
        layersource_objects = []
        for layersource in image_layers[image_layer]["layersources"]:
            layersource_object: Union[WMSLayerSource, FileLayerSource]
            try:
                # If not, the layersource should be specified in seperate parameters
                if "wms_server_url" in layersource:
                    layersource_object = WMSLayerSource(
                        wms_server_url=layersource["wms_server_url"],
                        wms_version=layersource.get("wms_version", "1.3.0"),
                        layernames=_str2list(layersource["wms_layernames"]),
                        layerstyles=_str2list(layersource.get("wms_layerstyles")),
                        bands=_str2intlist(layersource.get("bands", None)),
                        username=layersource.get("wms_username", None),
                        password=layersource.get("wms_password", None),
                        random_sleep=int(layersource.get("random_sleep", 0)),
                        wms_ignore_capabilities_url=_str2bool(
                            layersource.get("wms_ignore_capabilities_url", "False")
                        ),
                    )
                elif "path" in layersource:
                    path = Path(layersource["path"])
                    if not path.is_absolute():
                        # Resolve relative path based on layer_config_filepath.parent
                        path = layer_config_filepath.parent / layersource["path"]
                        path = path.resolve()
                    layersource_object = FileLayerSource(
                        path=path,
                        layernames=_str2list(layersource["layername"]),
                        bands=_str2intlist(layersource.get("bands", None)),
                    )
                else:
                    raise ValueError(
                        f"Invalid layersource, should be WMS or file: {layersource}"
                    )
            except Exception as ex:
                raise ValueError(
                    f"Missing parameter in image_layer {image_layer}, layersource "
                    f"{layersource}: {ex}"
                ) from ex
            layersource_objects.append(layersource_object)
        image_layers[image_layer]["layersources"] = layersource_objects

        # Read nb_concurrent calls param
        image_layers[image_layer]["nb_concurrent_calls"] = layer_config[
            image_layer
        ].getint("nb_concurrent_calls", fallback=1)

        # Check if a region of interest is specified as file or bbox
        image_layers[image_layer]["roi_filepath"] = layer_config[image_layer].getpath(
            "roi_filepath", fallback=None
        )
        bbox_tuple = None
        if layer_config.has_option(image_layer, "bbox"):
            bbox_list = layer_config[image_layer].getlist("bbox")
            bbox_tuple = (
                float(bbox_list[0]),
                float(bbox_list[1]),
                float(bbox_list[2]),
                float(bbox_list[3]),
            )
        image_layers[image_layer]["bbox"] = bbox_tuple

        # Check if the grid xmin and xmax are specified
        image_layers[image_layer]["grid_xmin"] = layer_config[image_layer].getfloat(
            "grid_xmin", fallback=0
        )
        image_layers[image_layer]["grid_ymin"] = layer_config[image_layer].getfloat(
            "grid_ymin", fallback=0
        )

        # Check if a image_pixels_ignore_border is specified
        image_layers[image_layer]["image_pixels_ignore_border"] = layer_config[
            image_layer
        ].getint("image_pixels_ignore_border", fallback=0)

        # Convert pixel_x_size and pixel_y_size to float
        image_layers[image_layer]["pixel_x_size"] = layer_config[image_layer].getfloat(
            "pixel_x_size"
        )
        image_layers[image_layer]["pixel_y_size"] = layer_config[image_layer].getfloat(
            "pixel_y_size"
        )

    return image_layers


def _prepare_train_label_infos(
    labelpolygons_pattern: Path,
    labellocations_pattern: Path,
    label_datasources: dict,
    image_layers: dict[str, dict],
) -> list[LabelInfo]:
    # Search for the files based on the file name patterns...
    label_infos: dict[str, LabelInfo] = {}
    if labelpolygons_pattern is not None or labellocations_pattern is not None:
        label_infos = {
            info.locations_path.resolve().as_posix(): info
            for info in _search_label_files(
                labelpolygons_pattern, labellocations_pattern
            )
        }

    # If there are label datasources configured in the project, process them as well.
    if label_datasources is not None:
        for label_ds_key in label_datasources:
            label_ds = label_datasources[label_ds_key]

            # Prepare polygons_path with backwards compatibility for "data_path"
            polygons_path = None
            if "polygons_path" in label_ds:
                polygons_path = label_ds["polygons_path"]
            elif "data_path" in label_ds:
                polygons_path = label_ds["data_path"]

            # If the label datasource was already found using pattern search, overrule
            # extra info rather than adding one.
            locations_path = Path(label_ds["locations_path"]).resolve().as_posix()
            if locations_path in label_infos:
                if polygons_path is not None:
                    label_infos[locations_path].polygons_path = Path(polygons_path)
                if label_ds.get("image_layer") is not None:
                    label_infos[locations_path].image_layer = label_ds["image_layer"]
                if label_ds.get("pixel_x_size") is not None:
                    label_infos[locations_path].pixel_x_size = label_ds["pixel_x_size"]
                if label_ds.get("pixel_y_size") is not None:
                    label_infos[locations_path].pixel_y_size = label_ds["pixel_y_size"]
            else:
                if polygons_path is None:
                    raise ValueError(f"polygons_path not specified for {label_ds}")

                # Add as new LabelInfo
                label_infos[locations_path] = LabelInfo(
                    locations_path=Path(label_ds["locations_path"]),
                    polygons_path=Path(polygons_path),
                    image_layer=label_ds["image_layer"],
                    pixel_x_size=label_ds.get("pixel_x_size"),
                    pixel_y_size=label_ds.get("pixel_y_size"),
                )

    # Check if the configured image_layer exists for all label_infos
    for label_info in label_infos.values():
        if label_info.image_layer not in image_layers:
            raise ValueError(
                f"invalid image_layer in {label_info}: not in {list(image_layers)}"
            )

    return list(label_infos.values())


def _search_label_files(
    labelpolygons_pattern: Path, labellocations_pattern: Path
) -> list[LabelInfo]:
    if not labelpolygons_pattern.parent.exists():
        raise ValueError(f"Label dir doesn't exist: {labelpolygons_pattern.parent}")
    if not labellocations_pattern.parent.exists():
        raise ValueError(f"Label dir doesn't exist: {labellocations_pattern.parent}")

    label_infos = []
    labelpolygons_pattern_searchpath = Path(
        str(labelpolygons_pattern).format(image_layer="*")
    )
    labelpolygons_paths = list(
        labelpolygons_pattern_searchpath.parent.glob(
            labelpolygons_pattern_searchpath.name
        )
    )
    labellocations_pattern_searchpath = Path(
        str(labellocations_pattern).format(image_layer="*")
    )
    labellocations_paths = list(
        labellocations_pattern_searchpath.parent.glob(
            labellocations_pattern_searchpath.name
        )
    )

    # Loop through all labellocation files
    for labellocations_path in labellocations_paths:
        tokens = _unformat(labellocations_path.stem, labellocations_pattern.stem)
        if "image_layer" not in tokens:
            raise ValueError(  # pragma: no cover
                f"image_layer token not found in {labellocations_path} using pattern "
                f"{labellocations_pattern}"
            )
        image_layer = tokens["image_layer"]

        # Look for the matching (= same image_layer) data file
        found = False
        for labelpolygons_path in labelpolygons_paths:
            tokens = _unformat(labelpolygons_path.stem, labelpolygons_pattern.stem)
            if "image_layer" not in tokens:
                raise ValueError(  # pragma: no cover
                    f"image_layer token not found in {labelpolygons_path} using "
                    f"pattern {labelpolygons_pattern}"
                )

            if tokens["image_layer"] == image_layer:
                found = True
                break

        if found is False:
            raise ValueError(
                f"no matching polygon data file found for {labellocations_path}"
            )
        label_infos.append(
            LabelInfo(
                locations_path=labellocations_path,
                polygons_path=labelpolygons_path,
                image_layer=image_layer,
            )
        )

    return label_infos


def _unformat(string: str, pattern: str) -> dict:
    regex = re.sub(r"{(.+?)}", r"(?P<_\1>.+)", pattern)
    regex_result = re.search(regex, string)
    if regex_result is not None:
        values = list(regex_result.groups())
        keys = re.findall(r"{(.+?)}", pattern)
        _dict = dict(zip(keys, values))
        return _dict
    else:
        raise ValueError(f"pattern {pattern} not found in {string}")


def _str2list(input: Optional[str]):
    if input is None:
        return None
    if isinstance(input, list):
        return input
    return [part.strip() for part in input.split(",")]


def _str2intlist(input: Optional[str]):
    if input is None:
        return None
    if isinstance(input, list):
        return input
    return [int(i.strip()) for i in input.split(",")]


def _str2bool(input: Optional[str]):
    if input is None:
        return None
    if isinstance(input, bool):
        return input
    return input.lower() in ("yes", "true", "false", "1")



"""Module with high-level operations to segment images."""

import csv
import datetime
import json
import logging
import multiprocessing
import shutil
import tempfile
import time
import traceback
from concurrent import futures
from pathlib import Path
from typing import Any, Optional

import geofileops as gfo
import keras.models
import numpy as np
import pandas as pd
import rasterio as rio
import rasterio.crs as rio_crs
import rasterio.plot as rio_plot
import tensorflow as tf

import orthoseg.lib.postprocess_predictions as postp
from orthoseg.util import general_util
from orthoseg.util.progress_util import ProgressLogger

# Get a logger...
logger = logging.getLogger(__name__)


def predict_dir(
    model: keras.models.Model,
    input_image_dir: Path,
    output_image_dir: Path,
    output_vector_path: Optional[Path],
    classes: list,
    min_probability: float = 0.5,
    postprocess: dict = {},
    border_pixels_to_ignore: int = 0,
    projection_if_missing: Optional[str] = None,
    input_mask_dir: Optional[Path] = None,
    batch_size: int = 16,
    evaluate_mode: bool = False,
    cancel_filepath: Optional[Path] = None,
    nb_parallel_postprocess: int = 1,
    max_prediction_errors: int = 100,
    force: bool = False,
    no_images_ok: bool = False,
):
    """Create a prediction for all the images in a directory.

    If evaluate_mode is False, the output folder(s) will contain:
        * the "raw" prediction for every image (if there are white pixels)
        * a geojson with the vectorized prediction, with a column "onborder"
          for each feature that is 1 if the feature is on the border of the
          tile, taking the border_pixels_to_ignore in account if applicable.
          This columns can be used to speed up union operations afterwards,
          because only features on the border of tiles need to be unioned.

    If evaluate_mode is True, the results will all be put in the root of the
    output folder, and the following files will be outputted:
        * the original image
        * the mask that was provided, if available
        * the "raw" prediction
        * a "cleaned" version of the prediction
    The files will in this case be prefixed with a number so they are ordered
    in a way that is interesting for evaluation. If a mask was available, this
    prefix will be the % overlap of the mask and the prediction. If no mask is
    available, the prefix is the % white pixels in the prediction.

    Args:
        model (Model): the model to use for the prediction
        input_image_dir (Pathlike): dir where the input images are located
        output_image_dir (Pathlike): dir where the output will be put
        output_vector_path (Pathlike): the path to write the vector output to
        classes (list): a list of the different class names. Mandatory
            if more than background + 1 class.
        min_probability (float): Minimum probability to consider a pixel being of a
            certain class. Default to 0.5.
        postprocess (dict, optional): specifies which postprocessing should be applied
            to the prediction. Default is {}, so no postprocessing.
        border_pixels_to_ignore: because the segmentation at the borders of the
            input images images is not as good, you can specify that x
            pixels need to be ignored
        input_mask_dir: optional dir where the mask images are located
        projection_if_missing: Normally the projection should be in the raster file. If
            it is not, you can explicitly specify one.
        batch_size: batch size to use while predicting. This must be choosen
            depending on the neural network architecture and available
            memory on you GPU.
        evaluate_mode: True to run in evaluate mode
        cancel_filepath: If the file in this path exists, processing stops asap
        nb_parallel_postprocess (int, optional): The number of parallel
            processes used to vectorize,... the predictions. If -1, all
            available CPU's are used. Defaults to 1.
        max_prediction_errors (int, optional): the maximum number of errors that is
            tolerated before stopping prediction. If -1, no limit. Defaults to 100.
        force: False to skip images that already have a prediction, true to
            ignore existing predictions and overwrite them
        no_images_ok (bool, optional): False to throw `ValueError`
            when no images available in the `input_image_dir`,
            True to return without error. Defaults to False.
    """
    # Init
    if not input_image_dir.exists():
        logger.warning(f"input_image_dir doesn't exist, so return: {input_image_dir}")
        return
    if output_vector_path is not None and output_vector_path.exists():
        logger.warning(f"output file exists already, so return: {output_vector_path}")
        return

    # Create tmp dir for this predict run
    tmp_dir = Path(tempfile.gettempdir()) / "orthoseg"
    tmp_dir.mkdir(parents=True, exist_ok=True)
    tmp_dir = Path(tempfile.mkdtemp(prefix=f"{Path(__file__).stem}_", dir=tmp_dir))
    logger.info(f"Start predict for input_image_dir: {input_image_dir}")

    # Eager and not eager prediction seems +- the same performance-wise
    # model.run_eagerly = False

    # If we are using evaluate mode, change the output dir...
    if evaluate_mode:
        output_image_dir = Path(str(output_image_dir) + "_eval")

    # Create the output dir's if they don't exist yet...
    for dir in [output_image_dir, tmp_dir]:
        if not dir.exists():
            dir.mkdir()

    # Write prediction config used, so it can be used for postprocessing
    prediction_config_path = output_image_dir / "prediction_config.json"
    with open(prediction_config_path, "w") as pred_conf_file:
        pred_conf: dict[str, Any] = {}
        pred_conf["border_pixels_to_ignore"] = border_pixels_to_ignore
        pred_conf["classes"] = classes
        json.dump(pred_conf, pred_conf_file)

    # Get list of all image files to process and to skip
    image_filepaths: list[Path] = []
    input_ext = [".png", ".tif", ".jpg"]
    for input_ext_cur in input_ext:
        image_filepaths.extend(input_image_dir.rglob("*" + input_ext_cur))
    image_filepaths = sorted(image_filepaths)
    nb_images = len(image_filepaths)

    if nb_images == 0:
        if no_images_ok:
            return
        raise ValueError(f"Found no {input_ext} images to predict in {input_image_dir}")
    logger.info(f"Found {nb_images} {input_ext} images to predict in {input_image_dir}")

    # If force is false, get list of all existing predictions
    # Getting the list once is way faster than checking file per file later on!
    images_done_log_filepath = output_image_dir / "images_done.txt"
    image_done_filenames = set()
    if force is False:
        # First read the listing files if they exists
        if images_done_log_filepath.exists():
            with images_done_log_filepath.open() as f:
                for filename in f:
                    image_done_filenames.add(filename.rstrip())
        if len(image_done_filenames) > 0:
            logger.info(
                f"Skip {len(image_done_filenames)} images that are already processed"
            )

    # Clear error file
    images_error_log_filepath = output_image_dir / "images_error.csv"
    images_error_log_filepath.unlink(missing_ok=True)

    # Write output to tmp files, so we know if the process completed correctly or not
    pred_tmp_output_path = None
    if output_vector_path is not None:
        pred_tmp_output_path = output_image_dir / f"{output_vector_path.stem}_tmp.gpkg"
        pred_tmp_output_lock_path = Path(f"{pred_tmp_output_path!s}.lock")
        # if lock file exists, remove it:
        if pred_tmp_output_lock_path.exists():
            pred_tmp_output_lock_path.unlink()

    # Loop through all files to process them
    nb_parallel_read = batch_size * 3
    if nb_parallel_postprocess == -1:
        nb_parallel_postprocess = multiprocessing.cpu_count()
    predict_queue: list[dict] = []
    nb_to_predict = nb_images
    nb_done = 0
    nb_errors = 0
    read_sleep_logged = False
    progress = None
    read_queue: dict[futures.Future, Path] = {}
    postp_queue: dict[futures.Future, Path] = {}
    write_queue: dict[futures.Future, Path] = {}
    image_id = -1
    last_image_reached = False

    # We don't want the postprocess workers to block the entire system,
    # so make them a bit nicer
    def init_postprocess_worker():
        general_util.setprocessnice(15)

    with (
        futures.ThreadPoolExecutor(nb_parallel_read) as read_pool,
        futures.ProcessPoolExecutor(
            nb_parallel_postprocess, initializer=init_postprocess_worker()
        ) as postprocess_pool,
        futures.ProcessPoolExecutor(max_workers=1) as write_pool,
    ):
        # Start looping.
        # If ready to stop, the code below will break
        perf_time_start = datetime.datetime.now()
        while True:
            # If we are ready, stop!
            if (
                last_image_reached is True
                and len(read_queue) == 0
                and len(postp_queue) == 0
                and len(write_queue) == 0
            ):
                break

            # If the cancel file exists, stop processing...
            if cancel_filepath is not None and cancel_filepath.exists():
                print()
                logger.info(f"Cancel file found, so stop: {cancel_filepath}")
                break

            # Init for new loop
            perfinfo: list[str] = []

            # Fill the read queue
            # -------------------
            while not last_image_reached and len(read_queue) <= nb_parallel_read:
                image_id += 1

                # Last image reached
                if image_id >= len(image_filepaths) - 1:
                    last_image_reached = True

                image_filepath = image_filepaths[image_id]

                # Check if the image has been processed already
                if force is False and image_filepath.name in image_done_filenames:
                    logger.debug(
                        "Predict for image has already been done before and force is "
                        f"False, so skip: {image_filepath.name}"
                    )
                    nb_to_predict -= 1
                    continue

                # Schedule file to be read
                read_future = read_pool.submit(
                    read_image,
                    image_filepath=image_filepath,
                    projection_if_missing=projection_if_missing,
                )
                read_queue[read_future] = image_filepath

            # Check read_queue for images read and add them to predict_queue
            # --------------------------------------------------------------
            while len(read_queue) > 0 and len(predict_queue) < batch_size:
                # Prepare the images that have been read for predicting
                futures_done = [
                    future for future in read_queue if future.done() is True
                ]
                for future in futures_done:
                    # predict_queue should contain maximum batch_size images!
                    if len(predict_queue) >= batch_size:
                        break

                    try:
                        # Get the result from the read
                        read_result = future.result()
                        image_filepath_read = read_result["image_filepath"]

                        # Prepare the filepath for the output
                        output_suffix = ".tif"
                        if evaluate_mode:
                            # In evaluate mode, put everyting in output base dir for
                            # easier comparison
                            output_image_pred_dir = output_image_dir

                            # Prepare complete filepath for image prediction
                            output_image_pred_path = (
                                output_image_dir / image_filepath_read.stem
                            )
                        else:
                            # If saving predictions to images for real, keep hierarchic
                            # structure if present
                            tmp_output_filepath = Path(
                                str(image_filepath_read).replace(
                                    str(input_image_dir), str(output_image_dir)
                                )
                            )
                            output_image_pred_dir = tmp_output_filepath.parent
                            output_image_pred_path = (
                                output_image_pred_dir
                                / f"{image_filepath_read.stem}_pred{output_suffix}"
                            )

                        predict_queue.append(
                            {
                                "input_image_filepath": image_filepath_read,
                                "output_pred_filepath": output_image_pred_path,
                                "output_image_pred_dir": output_image_pred_dir,
                                "image_crs": read_result["image_crs"],
                                "image_transform": read_result["image_transform"],
                                "image_data": read_result["image_data"],
                            }
                        )

                    finally:
                        # Remove from queue...
                        del read_queue[future]

                # If enough images in predict_queue or if the read queue is not full
                # stop this loop
                if (
                    len(predict_queue) == batch_size
                    or len(read_queue) < nb_parallel_read
                ):
                    break

                # Wait a bit for images to be read, then try finding images again
                if not read_sleep_logged and not last_image_reached:
                    logger.info("Wait for images to be read")
                    read_sleep_logged = True
                time.sleep(0.01)

            # If sufficient images in predict queue -> predict
            # ------------------------------------------------
            if len(predict_queue) == batch_size or (
                last_image_reached is True and len(predict_queue) > 0
            ):
                read_sleep_logged = False
                perf_time_now = datetime.datetime.now()
                perfinfo += f"waiting for read took {perf_time_now-perf_time_start}"
                perf_time_start = perf_time_now

                # Predict!
                # --------
                logger.debug(f"Start prediction for {len(predict_queue)} images")
                perf_time_start = datetime.datetime.now()
                curr_batch_image_list = [
                    batch_image_info["image_data"] for batch_image_info in predict_queue
                ]
                batch_image_arr = np.stack(curr_batch_image_list)
                batch_pred_arr = model.predict_on_batch(batch_image_arr)

                perf_time_now = datetime.datetime.now()
                perfinfo += f", predict took {perf_time_now-perf_time_start}"
                perf_time_start = perf_time_now

                # In tf > 2.1 a tf.tensor object is returned, but we want an ndarray
                if isinstance(batch_pred_arr, tf.Tensor):
                    arr = batch_pred_arr.numpy()  # pyright: ignore[reportOptionalCall]
                    batch_pred_arr = np.array(arr)
                else:
                    batch_pred_arr = np.array(batch_pred_arr)

                # Add predictions to postprocess queue
                # ------------------------------------
                logger.debug("Start post-processing")
                for batch_image_id, image_info in enumerate(predict_queue):
                    try:
                        # If not in evaluate mode... save to vector in background
                        if (
                            evaluate_mode is False
                            and output_vector_path is not None
                            and pred_tmp_output_path is not None
                        ):
                            # Prepare prediction array...
                            #   - convert to uint8 to reduce pickle size/time
                            image_pred_arr_uint8 = (
                                (batch_pred_arr[batch_image_id, :, :, :]) * 255
                            ).astype(np.uint8)
                            # Save to specific temp file to avoid locking issues.
                            name = f"{image_info['input_image_filepath'].stem}.gpkg"
                            prd_tmp_partial_output_file = tmp_dir / name
                            future = postprocess_pool.submit(
                                postp.polygonize_pred_multiclass_to_file,
                                image_pred_arr=image_pred_arr_uint8,
                                image_crs=image_info["image_crs"],
                                image_transform=image_info["image_transform"],
                                classes=classes,
                                output_vector_path=prd_tmp_partial_output_file,
                                min_probability=min_probability,
                                postprocess=postprocess,
                                border_pixels_to_ignore=border_pixels_to_ignore,
                                create_spatial_index=False,
                            )
                            postp_queue[future] = image_info["input_image_filepath"]

                        else:
                            # Saving the predictions as images at the moment only used
                            # for evaluate mode...
                            # TODO: would ideally be moved to the background
                            # processing as well to simplify code here...
                            postp.clean_and_save_prediction(
                                input_image_filepath=image_info["input_image_filepath"],
                                image_crs=image_info["image_crs"],
                                image_transform=image_info["image_transform"],
                                image_pred_arr=batch_pred_arr[batch_image_id],
                                output_dir=image_info["output_image_pred_dir"],
                                input_image_dir=input_image_dir,
                                input_mask_dir=input_mask_dir,
                                border_pixels_to_ignore=border_pixels_to_ignore,
                                min_probability=min_probability,
                                evaluate_mode=evaluate_mode,
                                classes=classes,
                                force=force,
                            )

                            # Write filepath to file with files that are done
                            with images_done_log_filepath.open(
                                "a+"
                            ) as image_donelog_file:
                                image_donelog_file.write(
                                    f"{image_info['input_image_filepath'].name}\n"
                                )
                    except Exception as ex:
                        nb_errors += 1
                        image_path = image_info["input_image_filepath"]
                        _handle_error(image_path, ex, images_error_log_filepath)

                # Reset variable for next batch
                predict_queue = []

                # Collect performance debugging info
                perf_time_now = datetime.datetime.now()
                perfinfo += (
                    f", scheduling postprocessings took {perf_time_now-perf_time_start}"
                )
                perf_time_start = perf_time_now

            # Check postp_queue for completed postprocessings
            # -----------------------------------------------
            postp_sleep_logged = False
            while len(postp_queue) > 0:
                # If not at last file, get results from all futures that are
                # done, if at last file, wait till all are done
                futures_done = [
                    future for future in postp_queue if future.done() is True
                ]
                for future in futures_done:
                    # Get the result from the polygonization
                    image_path = postp_queue[future]
                    try:
                        # Get the result (= exception when something went wrong)
                        result = future.result()
                        logger.debug(f"result for {postp_queue[future].name}: {result}")

                        name = f"{image_path.stem}.gpkg"
                        partial_vector_path = tmp_dir / name
                        write_future = write_pool.submit(
                            _write_vector_result,
                            image_path=image_path,
                            partial_vector_path=partial_vector_path,
                            vector_output_path=pred_tmp_output_path,
                            images_done_log_filepath=images_done_log_filepath,
                        )
                        write_queue[write_future] = image_path
                    except ImportError as ex:
                        raise ex
                    except Exception as ex:
                        nb_errors += 1
                        _handle_error(image_path, ex, images_error_log_filepath)

                    finally:
                        # Remove from queue...
                        del postp_queue[future]

                # Wait till number below thresshold to avoid huge waiting
                # list (and memory issues)
                if len(postp_queue) > nb_parallel_postprocess * 2:
                    if postp_sleep_logged is False:
                        logger.info(
                            "Postprocessing takes longer than prediction, so wait"
                        )
                        postp_sleep_logged = True
                    time.sleep(0.01)
                else:
                    # No need to wait (anymore)...
                    if postp_sleep_logged is True:
                        logger.info("Waited enough for postprocessing to catch up...")

                    perf_time_now = datetime.datetime.now()
                    perfinfo += (
                        f", after postproces: took {perf_time_now-perf_time_start}"
                    )
                    perf_time_start = perf_time_now
                    break

            # Check write_queue for completed write operations
            # ------------------------------------------------
            write_sleep_logged = False
            while len(write_queue) > 0:
                # If not at last file, get results from all futures that are
                # done, if at last file, wait till all are done
                futures_done = [
                    future for future in write_queue if future.done() is True
                ]
                for future in futures_done:
                    # Get the result from the write
                    try:
                        # Get the result (= exception when something went wrong)
                        result = future.result()
                    except Exception as ex:
                        nb_errors += 1
                        image_path = write_queue[future]
                        _handle_error(image_path, ex, images_error_log_filepath)

                    finally:
                        # Remove from queue...
                        del write_queue[future]
                        nb_done += 1

                # Wait till number below thresshold to avoid huge waiting list (and
                # memory issues)
                if len(write_queue) > nb_parallel_postprocess * 2:
                    if write_sleep_logged is False:
                        logger.info("Writing takes longer than prediction, so wait")
                        write_sleep_logged = True
                    time.sleep(0.01)
                else:
                    # No need to wait (anymore)...
                    if write_sleep_logged is True:
                        logger.info("Waited enough for writing to catch up...")

                    break

            # Prepare for next loop
            # ---------------------
            # Log the progress and prediction speed
            if len(perfinfo) > 0:
                logger.debug(perfinfo)
            if progress is not None:
                progress.update(nb_steps_done=nb_done, nb_steps_total=nb_to_predict)
            # Init progress only when some imags were already processed, as the
            # first are very slow.
            if progress is None and nb_done > 0:
                progress = ProgressLogger(
                    message=f"predict to {output_image_dir.parent.name}/{output_image_dir.name}",  # noqa: E501
                    nb_steps_total=nb_to_predict,
                    nb_steps_done=nb_done,
                )

            # If max number errors reached, stop processings
            if max_prediction_errors >= 0 and nb_errors >= max_prediction_errors:
                break

        # If errors occured, raise error
        if images_error_log_filepath.exists():
            errors = pd.read_csv(
                images_error_log_filepath, usecols=["filename", "error"]
            ).to_html(justify="left", index=False)
            raise Exception(f"Error(s) occured while predicting:\n{errors}")

        # If all images were processed, rename to real output file + cleanup
        if (
            last_image_reached is True
            and output_vector_path is not None
            and pred_tmp_output_path is not None
            and pred_tmp_output_path.exists()
        ):
            output_vector_path.parent.mkdir(parents=True, exist_ok=True)
            gfo.create_spatial_index(pred_tmp_output_path, exist_ok=True)
            gfo.move(pred_tmp_output_path, output_vector_path)
            gfo.rename_layer(output_vector_path, output_vector_path.stem)
            shutil.rmtree(tmp_dir, ignore_errors=True)
            shutil.rmtree(output_image_dir)


def _write_vector_result(
    image_path: Path,
    partial_vector_path: Path,
    vector_output_path: Optional[Path],
    images_done_log_filepath: Path,
):
    # Copy the result to the main vector output file
    if vector_output_path is not None:
        if partial_vector_path.exists():
            gfo.append_to(
                src=partial_vector_path,
                dst=vector_output_path,
                dst_layer=vector_output_path.stem,
                create_spatial_index=False,
            )
            gfo.remove(partial_vector_path)

    # Write filepath to file with files that are done
    with images_done_log_filepath.open("a+") as image_donelog_file:
        image_donelog_file.write(image_path.name + "\n")


def _handle_error(image_path: Path, ex: Exception, log_path: Path):
    # Print exception + trace
    exception_trace = traceback.format_exc()
    exception_trace_print = exception_trace.replace("\n", "\n\t")
    logger.error(f"Error postprocessing pred for {image_path}: {exception_trace_print}")

    # Write error to error log file
    first_error = False
    if not log_path.exists():
        first_error = True
    with log_path.open("a+") as log_file:
        if first_error:
            log_file.write("filename,error,traceback\n")
        writer = csv.writer(log_file)
        exception_trace_csv = exception_trace.replace("\n", "\\n")
        fields = [image_path.name, ex, exception_trace_csv]
        writer.writerow(fields)


def read_image(
    image_filepath: Path, projection_if_missing: Optional[str] = None
) -> dict:
    """Read image file.

    Args:
        image_filepath (Path): file path.
        projection_if_missing (Optional[str], optional): _description_.
            Defaults to None.

    Returns:
        dict: _description_
    """
    # Read input file
    # Because sometimes a read seems to fail, retry up to 3 times...
    retry_count = 0
    while True:
        try:
            with rio.open(str(image_filepath)) as image_ds:
                # Read geo info
                image_crs = image_ds.profile["crs"]
                image_transform = image_ds.transform

                # Read pixels + change from (channels, width, height) to
                # (width, height, channels) + normalize to between 0 and 1
                image_data = image_ds.read()
                image_data = rio_plot.reshape_as_image(image_data)
                image_data = image_data / 255.0

            # Read worked, so jump out of the loop...
            break
        except Exception as ex:
            retry_count += 1
            logger.warning(f"Read failed, retry nb {retry_count} for {image_filepath}")
            if retry_count >= 3:
                message = f"Read failed {retry_count} times for {image_filepath}: {ex}"
                logger.error(message)
                raise RuntimeError(message)

    # The read was successfull, now check if there was a projection in the
    # file and/or if one was provided
    if image_crs is None:
        if projection_if_missing is not None:
            image_crs = rio_crs.CRS.from_string(projection_if_missing)
        else:
            message = (
                f"Image has no proj and projection_if_missing is None: {image_filepath}"
            )
            logger.error(message)
            raise ValueError(message)

    # Now return the result
    result = {
        "image_data": image_data,
        "image_crs": image_crs,
        "image_transform": image_transform,
        "image_filepath": image_filepath,
    }
    return result



"""
Tests for functionalities in orthoseg.lib.postprocess_predictions.
"""

import filecmp
import math
import os
from pathlib import Path
from typing import Optional, Union

import geofileops as gfo
import geopandas as gpd
import pytest
from PIL import Image
from shapely import geometry as sh_geom

from tests import test_helper

# Make hdf5 version warning non-blocking
os.environ["HDF5_DISABLE_VERSION_CHECK"] = "1"

from orthoseg.helpers import config_helper
from orthoseg.lib import prepare_traindatasets as prep_traindata
from orthoseg.lib.prepare_traindatasets import ValidationError
from tests.test_helper import TestData


def _prepare_locations_file(
    tmp_path, locations: Optional[Union[dict, gpd.GeoDataFrame]]
) -> Path:
    locations_path = tmp_path / "locations.gpkg"
    if locations is None:
        locations = TestData.locations_gdf
    if isinstance(locations, dict):
        locations = gpd.GeoDataFrame(locations, crs="EPSG:31370")

    gfo.to_file(locations, locations_path)
    return locations_path


def _prepare_polygons_file(
    tmp_path, polygons: Optional[Union[dict, gpd.GeoDataFrame]]
) -> Path:
    polygons_path = tmp_path / "polygons.gpkg"
    if polygons is None:
        polygons = TestData.polygons_gdf
    if isinstance(polygons, dict):
        polygons = gpd.GeoDataFrame(polygons, crs="EPSG:31370")

    gfo.to_file(polygons, polygons_path)
    return polygons_path


def _prepare_labelinfos(
    tmp_path,
    locations: Optional[Union[dict, gpd.GeoDataFrame]] = None,
    polygons: Optional[Union[dict, gpd.GeoDataFrame]] = None,
) -> list[prep_traindata.LabelInfo]:
    locations_path = _prepare_locations_file(tmp_path, locations)
    polygons_path = _prepare_polygons_file(tmp_path, polygons)

    label_info = prep_traindata.LabelInfo(
        locations_path=locations_path,
        polygons_path=polygons_path,
        image_layer="BEFL-2019",
    )
    return [label_info]


@pytest.mark.parametrize(
    "geometry, traindata_type, expected_len_locations",
    [
        (TestData.location, "train", 4),
        (None, "validation", 2),
        (sh_geom.Polygon(), "validation", 2),
    ],
)
def test_prepare_labeldata_locations(geometry, traindata_type, expected_len_locations):
    # Prepare test data
    # Add some extra data to make sure it also works for multiple rows
    # Prepare test data, by wrapping the parametrized invalid test data by proper data
    # to make sure the checks work on multiple rows,...
    locations_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.location, geometry, geometry, TestData.location],
            "traindata_type": ["train", traindata_type, traindata_type, "validation"],
            "path": "/tmp/locations.gdf",
        },
        crs="EPSG:31370",
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=locations_gdf,
        polygons_gdf=TestData.polygons_gdf,
    )

    # Test!
    labeldata = prep_traindata.prepare_labeldata(
        label_infos=[label_info],
        classes=TestData.classes,
        labelname_column="classname",
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )
    labellocations_gdf, labelpolygons_gdf = labeldata[0]

    assert len(labelpolygons_gdf) == len(TestData.polygons_gdf)
    assert len(labellocations_gdf) == expected_len_locations


@pytest.mark.parametrize(
    "expected_errors, geometries, traindata_types",
    [
        (["Invalid geometry "] * 4, [TestData.location_invalid] * 4, None),
        (
            "Invalid traindata_type ",
            None,
            ["train", "train", "validation", "invalid_traindata_type"],
        ),
        (
            ["Location geometry skewed or too small "] * 4,
            [TestData.location.buffer(-5)] * 4,
            None,
        ),
        ("No labellocations with traindata_type == 'validation' ", None, ["train"] * 4),
        ("No labellocations with traindata_type == 'train' ", None, ["validation"] * 4),
    ],
)
def test_prepare_labeldata_locations_invalid(
    expected_errors, geometries, traindata_types
):
    # Prepare test data, by wrapping the parametrized invalid test data by proper data
    # to make sure the checks work on multiple rows,...
    if isinstance(expected_errors, str):
        expected_errors = [expected_errors]
    if geometries is None:
        geometries = [TestData.location] * 4
    if traindata_types is None:
        traindata_types = ["train", "train", "train", "validation"]
    locations_gdf = gpd.GeoDataFrame(
        data={
            "geometry": geometries,
            "traindata_type": traindata_types,
            "path": "/tmp/locations.gdf",
        },
        crs="EPSG:31370",
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=locations_gdf,
        polygons_gdf=TestData.polygons_gdf,
    )

    # Test!
    with pytest.raises(ValidationError, match="Errors found in label data") as ex:
        _ = prep_traindata.prepare_labeldata(
            label_infos=[label_info],
            classes=TestData.classes,
            labelname_column="classname",
            image_pixel_x_size=TestData.image_pixel_x_size,
            image_pixel_y_size=TestData.image_pixel_y_size,
            image_pixel_width=TestData.image_pixel_width,
            image_pixel_height=TestData.image_pixel_height,
        )

    assert ex.value.errors is not None
    assert len(ex.value.errors) == len(expected_errors)
    for idx, expected_error in enumerate(expected_errors):
        assert ex.value.errors[idx].startswith(expected_error)


@pytest.mark.parametrize(
    "geometry, classname, expected_len_polygons",
    [
        [TestData.polygon, "testlabel1", 4],
        [None, "testlabel1", 2],
        [sh_geom.Polygon(), "testlabel1", 2],
    ],
)
def test_prepare_labeldata_polygons(geometry, classname, expected_len_polygons):
    # Prepare test data
    # Add some extra data to make sure it also works for multiple rows
    polygons_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, geometry, geometry, TestData.polygon],
            "classname": ["testlabel1", classname, classname, "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="EPSG:31370",
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=TestData.locations_gdf,
        polygons_gdf=polygons_gdf,
    )

    # Test!
    labeldata = prep_traindata.prepare_labeldata(
        label_infos=[label_info],
        classes=TestData.classes,
        labelname_column="classname",
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )
    labellocations_gdf, labelpolygons_gdf = labeldata[0]

    assert len(labelpolygons_gdf) == expected_len_polygons
    assert len(labellocations_gdf) == len(TestData.locations_gdf)


@pytest.mark.parametrize(
    "expected_error, geometry, classname",
    [
        ("Invalid geometry ", TestData.polygon_invalid, "testlabel1"),
        ("Invalid classname ", TestData.polygon, "unknown"),
    ],
)
def test_prepare_labeldata_polygons_invalid(expected_error, geometry, classname):
    # Prepare test data, by wrapping the parametrized invalid test data by proper data
    # to make sure the checks work on multiple rows,...
    polygons_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, geometry, geometry, TestData.polygon],
            "classname": ["testlabel1", classname, classname, "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="EPSG:31370",
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=TestData.locations_gdf,
        polygons_gdf=polygons_gdf,
    )

    # Test!
    with pytest.raises(ValidationError, match="Errors found in label data") as ex:
        _ = prep_traindata.prepare_labeldata(
            label_infos=[label_info],
            classes=TestData.classes,
            labelname_column="classname",
            image_pixel_x_size=TestData.image_pixel_x_size,
            image_pixel_y_size=TestData.image_pixel_y_size,
            image_pixel_width=TestData.image_pixel_width,
            image_pixel_height=TestData.image_pixel_height,
        )

    assert ex.value.errors is not None
    assert len(ex.value.errors) == 2
    assert ex.value.errors[0].startswith(expected_error)
    assert ex.value.errors[1].startswith(expected_error)


def test_prepare_labeldata_polygons_columnname_backw_compat():
    # Test bacwards compatibility for old label column name
    # Prepare test data
    polygons_gdf = gpd.GeoDataFrame(
        data={
            "geometry": [TestData.polygon, TestData.polygon],
            "label_name": ["testlabel1", "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="EPSG:31370",
    )
    label_info = prep_traindata.LabelInfo(
        locations_path=Path("locations"),
        polygons_path=Path("polygons"),
        image_layer="BEFL-2019",
        locations_gdf=TestData.locations_gdf,
        polygons_gdf=polygons_gdf,
    )

    # Test!
    labeldata = prep_traindata.prepare_labeldata(
        label_infos=[label_info],
        classes=TestData.classes,
        labelname_column="test_columnname",
        image_pixel_x_size=TestData.image_pixel_x_size,
        image_pixel_y_size=TestData.image_pixel_y_size,
        image_pixel_width=TestData.image_pixel_width,
        image_pixel_height=TestData.image_pixel_height,
    )
    locations_gdf, polygons_to_burn_gdf = labeldata[0]

    assert len(polygons_to_burn_gdf) == 2


image_1 = "150100_150100_150228_150228_512_512_OMWRGB19VL.png"
image_2 = "150200_150200_150328_150328_512_512_OMWRGB19VL.png"
image_3 = "150300_150300_150428_150428_512_512_OMWRGB19VL.png"


# @pytest.mark.skipif(os.name == "nt", reason="crashes on windows")
@pytest.mark.parametrize(
    "copy_images, match, mismatch, error, filesize_kb",
    [
        ("add_one", [image_1, image_2], [], [image_3], (1, 1, 2)),
        ("remove_one", [image_1], [], [image_2, image_3], (1, None, None)),
        ("copy_all", [image_1, image_2], [], [image_3], (1, 1, None)),
        ("copy_none", [], [], [image_1, image_2, image_3], (None, None, 2)),
    ],
    ids=[
        "add one new location",
        "remove one location",
        "all the same locations",
        "all different locations",
    ],
)
def test_prepare_traindatasets(
    tmp_path, copy_images: str, match, mismatch, error, filesize_kb
):
    # Prepare test data
    classes = TestData.classes
    image_layers_config_path = test_helper.sampleprojects_dir / "imagelayers.ini"
    image_layers = config_helper._read_layer_config(image_layers_config_path)
    label_infos = _prepare_labelinfos(tmp_path)

    training_dir = tmp_path / "training"
    previous_training_version_dir = training_dir / "01"
    traindata_types = ["train", "validation", "test"]

    for traindata_type in traindata_types:
        dir = previous_training_version_dir / traindata_type
        (dir / "image").mkdir(parents=True, exist_ok=True)
        (dir / "mask").mkdir(parents=True, exist_ok=True)
        for image in [image_1, image_2]:
            img = Image.new(
                mode="RGB",
                size=(TestData.image_pixel_width, TestData.image_pixel_height),
            )
            img.save(dir / "image" / image, "png")
            (dir / "image" / image.replace(".png", ".pgw")).touch()

    locations_gdf = _create_locations_dataframe(copy_images=copy_images)

    # Run prepare traindatasets
    label_infos = _prepare_labelinfos(tmp_path=tmp_path, locations=locations_gdf)
    output_dir, _ = prep_traindata.prepare_traindatasets(
        label_infos=label_infos,
        classes=classes,
        image_layers=image_layers,
        training_dir=training_dir,
    )

    # Check if files exist in new output folder and if they have the correct size
    # if size = 1kb then copied else if size = 2kb downloaded
    for traindata_type in traindata_types:
        images_to_compare = [image_1, image_2, image_3]
        assert filecmp.cmpfiles(
            previous_training_version_dir / traindata_type / "image",
            output_dir / traindata_type / "image",
            images_to_compare,
            shallow=False,
        ) == (match, mismatch, error)

        for i, image in enumerate(images_to_compare):
            image_path = output_dir / traindata_type / "image" / image
            if filesize_kb[i] is None:
                assert not image_path.exists()
            else:
                assert image_path.exists()
                assert math.ceil(image_path.stat().st_size / 1024) == filesize_kb[i]


def _create_locations_dataframe(copy_images: str) -> gpd.GeoDataFrame:
    def _create_location(crs_xmin: int, crs_ymin: int):
        return sh_geom.box(
            crs_xmin,
            crs_ymin,
            crs_xmin + (TestData.image_pixel_width * TestData.image_pixel_x_size),
            crs_ymin + (TestData.image_pixel_height * TestData.image_pixel_y_size),
        )

    location_1 = _create_location(crs_xmin=150100, crs_ymin=150100)
    location_2 = _create_location(crs_xmin=150200, crs_ymin=150200)
    location_3 = _create_location(crs_xmin=150300, crs_ymin=150300)

    if copy_images == "add_one":
        locations_gdf = gpd.GeoDataFrame(
            {
                "geometry": [
                    location_1,
                    location_1,
                    location_1,
                    location_2,
                    location_2,
                    location_2,
                    location_3,
                    location_3,
                    location_3,
                ],
                "traindata_type": [
                    "train",
                    "validation",
                    "test",
                    "train",
                    "validation",
                    "test",
                    "train",
                    "validation",
                    "test",
                ],
                "path": "/tmp/locations.gdf",
            },
            crs="epsg:31370",
        )

    elif copy_images == "remove_one":
        locations_gdf = gpd.GeoDataFrame(
            {
                "geometry": [location_1, location_1, location_1],
                "traindata_type": ["train", "validation", "test"],
                "path": "/tmp/locations.gdf",
            },
            crs="epsg:31370",
        )
    elif copy_images == "copy_all":
        locations_gdf = gpd.GeoDataFrame(
            {
                "geometry": [
                    location_1,
                    location_1,
                    location_1,
                    location_2,
                    location_2,
                    location_2,
                ],
                "traindata_type": [
                    "train",
                    "validation",
                    "test",
                    "train",
                    "validation",
                    "test",
                ],
                "path": "/tmp/locations.gdf",
            },
            crs="epsg:31370",
        )
    elif copy_images == "copy_none":
        locations_gdf = gpd.GeoDataFrame(
            {
                "geometry": [
                    location_3,
                    location_3,
                    location_3,
                ],
                "traindata_type": ["train", "validation", "test"],
                "path": "/tmp/locations.gdf",
            },
            crs="epsg:31370",
        )

    return locations_gdf



"""OrthoSeg makes it easy to train neural networks to segment orthophotos."""

from pathlib import Path

# ruff: noqa: F401
from orthoseg.load_images import load_images
from orthoseg.postprocess import postprocess
from orthoseg.predict import predict
from orthoseg.train import train
from orthoseg.validate import validate


def _get_version():
    version_path = Path(__file__).resolve().parent / "version.txt"
    with open(version_path) as file:
        return file.readline()


__version__ = _get_version()



"""
Tests for functionalities in orthoseg.model.model_factory.
"""

from typing import Optional

import pytest

from orthoseg.model import model_factory


@pytest.mark.parametrize(
    "architecture, input_width, input_height, expected_error",
    [
        ["test+linknet", 100, 100, "for decoder linknet"],
        ["test+linknet", 256, 256, None],
        ["test+pspnet", 144, 144, None],
        ["test+unet", 256, 256, None],
    ],
)
def test_check_image_size(
    architecture: str,
    input_width: int,
    input_height: int,
    expected_error: Optional[str],
):
    if expected_error is not None:
        with pytest.raises(ValueError, match=expected_error):
            model_factory.check_image_size(
                architecture=architecture,
                input_width=input_width,
                input_height=input_height,
            )
    else:
        model_factory.check_image_size(
            architecture=architecture,
            input_width=input_width,
            input_height=input_height,
        )


@pytest.mark.parametrize(
    "architecture, input_width, input_height",
    [
        ["mobilenetv2+linknet", 256, 256],
        ["mobilenetv2+pspnet", 144, 144],
        ["mobilenetv2+unet", 256, 256],
    ],
)
def test_get_compile_save_load_model(
    tmp_path, architecture: str, input_width: int, input_height: int
):
    # Get model
    model = model_factory.get_model(
        architecture=architecture,
        input_width=input_width,
        input_height=input_height,
        nb_classes=5,
    )
    assert model is not None

    # Compile model
    model = model_factory.compile_model(
        model,
        optimizer="adam",
        optimizer_params={"learning_rate": 0.0001},
        loss="categorical_crossentropy",
    )
    assert model is not None

    # Now save model
    model_path = tmp_path / f"{architecture}.hdf5"
    model.save(str(model_path))
    model = None

    # Load model again
    model = model_factory.load_model(model_path)
    assert model is not None


@pytest.mark.parametrize("architecture", ["mobilenetv2+unknown"])
def test_get_model_unknown_decoder(architecture: str):
    with pytest.raises(ValueError, match="Unknown decoder architecture:"):
        _ = model_factory.get_model(
            architecture=architecture,
        )


@pytest.mark.parametrize("architecture", ["unknown+unet"])
def test_get_model_unknown_encoder(architecture: str):
    # Error is raised by segmentation_models library
    with pytest.raises(ValueError, match="No such model"):
        _ = model_factory.get_model(
            architecture=architecture,
        )



"""Configuration setting to build the orthoseg package."""

import setuptools

with open("README.md") as fh:
    long_description = fh.read()

with open("orthoseg/version.txt") as file:
    version = file.readline()

setuptools.setup(
    name="orthoseg",
    version=version,
    author="Pieter Roggemans",
    author_email="pieter.roggemans@gmail.com",
    description="Package to make it easier to segment orthophotos.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/orthoseg/orthoseg",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        "gdown",
        "geofileops>=0.6,<0.10",
        "geopandas>=0.12,<1.1",
        "matplotlib",
        "numpy<2",
        "owslib",
        "pillow",
        "pycron",
        "pygeoops>=0.2,<0.5",
        "rasterio",
        "segmentation-models>=1.0,<1.1",
        "shapely>=2",
        "simplification",
        "tensorflow>=2.8,<2.11",
    ],
    entry_points="""
            [console_scripts]
            orthoseg_load_images=orthoseg.load_images:main
            orthoseg_validate=orthoseg.validate:main
            orthoseg_train=orthoseg.train:main
            orthoseg_predict=orthoseg.predict:main
            orthoseg_postprocess=orthoseg.postprocess:main
            osscriptrunner=orthoseg.scriptrunner:main
            orthoseg_load_sampleprojects=orthoseg.load_sampleprojects:main
            """,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)
